"""
The Config (:mod:`harp.config`) module provides tools to configure *Applications*.

*Applications* are standard python packages that contains specific definitions to be able to register services and
configure themselves. Every feature of Harp is built as an application, and you will want to write applications if you
want to extend harp's functionality for specific needs.


Contents
--------

"""

__title__ = "Config"

from .application import Application
from .config import Config
from .settings import (
    BaseSetting,
    Definition,
    DisableableBaseSettings,
    DisabledSettings,
    FromFileSetting,
    Lazy,
    asdict,
    settings_dataclass,
)

__all__ = [
    "Application",
    "BaseSetting",
    "Config",
    "Definition",
    "DisableableBaseSettings",
    "DisabledSettings",
    "FromFileSetting",
    "Lazy",
    "asdict",
    "settings_dataclass",
]



import binascii
from base64 import b64decode
from functools import cached_property
from typing import TYPE_CHECKING, MutableMapping, cast

from multidict import CIMultiDict, CIMultiDictProxy, MultiDictProxy

from ..utils.collections import MultiChainMap
from .typing import BaseHttpMessage, HttpRequestBridge
from .utils.cookies import parse_cookie


class HttpRequest(BaseHttpMessage):
    kind = "request"

    def __init__(self, impl: HttpRequestBridge):
        super().__init__()

        self._impl = impl
        self._body = []
        self._closed = False

    @cached_property
    def server_ipaddr(self) -> str:
        """Get the server IP address from the implementation bridge."""
        return self._impl.get_server_ipaddr()

    @cached_property
    def server_port(self) -> int:
        """Get the server port from the implementation bridge."""
        return self._impl.get_server_port()

    @cached_property
    def method(self) -> str:
        """Get the HTTP method from the implementation bridge."""
        return self._impl.get_method()

    @cached_property
    def path(self) -> str:
        """Get the HTTP path from the implementation bridge."""
        return self._impl.get_path()

    @cached_property
    def query(self) -> MultiDictProxy:
        """Get the query string from the implementation bridge, as a read only proxy."""
        return MultiDictProxy(self._impl.get_query())

    @cached_property
    def headers(self) -> CIMultiDictProxy:
        return CIMultiDictProxy(self._impl.get_headers())

    # body, content, raw_body

    @cached_property
    def cookies(self) -> dict:
        """Parse and returns cookies from headers."""
        cookies = {}
        for header in self.headers.getall("cookie", []):
            cookies.update(parse_cookie(header))
        return cookies

    @cached_property
    def basic_auth(self) -> tuple[str, str] | None:
        """Parse and returns basic auth from headers."""
        for header in self.headers.getall("authorization", []):
            try:
                _type, _auth = header.split(" ", 1)
            except ValueError:
                continue

            if _type.lower() == "basic":
                try:
                    user, passwd = b64decode(_auth).decode("utf-8").split(":", 1)
                except (binascii.Error, ValueError):
                    continue

                return user, passwd

    @cached_property
    def body(self) -> bytes:
        """Returns the previously read body of the request. Raises a RuntimeError if the body has not been read yet,
        you must await the `read()` asynchronous method first, which cannot be done here because properties are
        synchronous, so we let the user explicitely call it before."""
        if not self._closed:
            raise RuntimeError("Request body has not been read yet, please await `read()` first.")
        return b"".join(self._body)

    async def join(self):
        """Read all chunks from request. We may want to be able to read partial body later, but for now it's all or
        nothing. This method does nothing if the body has already been read."""
        if not self._closed:
            async for chunk in self._impl.stream():
                self._body.append(chunk)
            self._closed = True


class WrappedHttpRequest(HttpRequest):
    def __init__(self, wrapped: HttpRequest, /):
        # we do not need to call super, but we don't want linters to complain
        if TYPE_CHECKING:
            super().__init__(wrapped._impl)

        self._wrapped = wrapped
        self._headers = CIMultiDict()

    @property
    def _impl(self):
        return self._wrapped._impl

    @property
    def headers(self) -> MutableMapping:
        return MultiChainMap(self._headers, cast(MutableMapping, self._wrapped.headers))

    def __getattr__(self, item):
        return getattr(self._wrapped, item)



from typing import Optional

from whistle import AsyncEventDispatcher, IDispatchedEvent, IEvent, IListener

from harp import get_logger

logger = get_logger(__name__)
__title__ = "Event Dispatcher"


class LoggingAsyncEventDispatcher(AsyncEventDispatcher):
    """
    Adds logging to AsyncEventDispatcher, should probably go into whistle 2.x (with a bit of reengineering).

    todo: pass logger or logger name to constructor, choose logging level
    todo: add check for non-coroutines listeners which is wrong but leads to an undecypherable error message
    """

    def add_listener(self, event_id: str, listener: IListener, /, *, priority: int = 0):
        logger.info(f"👂 [Add] {event_id} ({listener})")
        super().add_listener(event_id, listener, priority=priority)

    async def adispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent:
        logger.info(f"⚡ [Dispatch] {event_id} ({type(event).__name__})")
        try:
            return await super().adispatch(event_id, event)
        except Exception as e:
            logger.error(f"⚡ [Error] {event_id} ({type(event).__name__}) failed: {e}")
            raise

    async def _adispatch(self, listeners, event):
        for listener in listeners:
            logger.debug(f"⚡ [DispatchOne] listener: {listener}")
            await listener(event)
            if event.propagation_stopped:
                break



from harp.config import Application
from harp.config.events import FactoryBuildEvent

from .settings import SentrySettings


class SentryApplication(Application):
    settings_namespace = "sentry"
    settings_type = SentrySettings

    @classmethod
    def defaults(cls, settings=None):
        return settings if settings is not None else {"dsn": None}

    async def on_build(self, event: FactoryBuildEvent):
        if not self.settings.dsn:
            return

        import sentry_sdk
        from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

        sentry_sdk.init(
            dsn=self.settings.dsn,
            traces_sample_rate=1.0,
        )

        event.kernel = SentryAsgiMiddleware(event.kernel)



from functools import cached_property

from multidict import MultiDict, MultiDictProxy

from .typing import BaseHttpMessage


class HttpError(BaseHttpMessage):
    kind = "error"

    def __init__(self, message: str, /, *, exception: Exception = None):
        super().__init__()
        self.message = message
        self.exception = exception

    @cached_property
    def headers(self) -> MultiDictProxy:
        return MultiDictProxy(MultiDict())

    @cached_property
    def body(self) -> bytes:
        if self.exception:
            msg = str(self.exception)
            out = type(self.exception).__name__
            if msg:
                out += f": {msg}"
            return out.encode()
        return self.message.encode()



from config.common import Configuration
from rodi import Container, Services
from whistle import Event

from harp.controllers import ProxyControllerResolver


class FactoryBindEvent(Event):
    """
    «Factory Bind» event happens before the service container is resolved.

    It is the right place to define services that may not be resolvable yet (because their dependencies may, or may not,
    be defined already).

    """

    name = "harp.config.bind"

    def __init__(self, container: Container, settings: Configuration):
        self.container = container
        self.settings = settings


EVENT_FACTORY_BIND = FactoryBindEvent.name


class FactoryBoundEvent(Event):
    """
    «Factory Bound» event happens after the service container is resolved, before the kernel creation.

    It is the right place to setup things that needs the container to be resolved (thus, you get a `provider` instead
    of a `container`).

    """

    name = "harp.config.bound"

    def __init__(self, provider: Services, resolver: ProxyControllerResolver):
        self.provider = provider
        self.resolver = resolver


EVENT_FACTORY_BOUND = FactoryBoundEvent.name


class FactoryBuildEvent(Event):
    """
    «Factory Build» event happens after the asgi kernel is created.

    It is the right place to decorate the kernel with middlewares.
    """

    name = "harp.config.build"

    def __init__(self, kernel, binds):
        self.kernel = kernel
        self.binds = binds


EVENT_FACTORY_BUILD = FactoryBuildEvent.name



from httpx import AsyncClient

from harp import get_logger
from harp.config import Application
from harp.config.events import FactoryBindEvent

from .client import AsyncHttpClient
from .settings import HttpClientSettings

logger = get_logger(__name__)


class HttpClientApplication(Application):
    settings_namespace = "http_client"
    settings_type = HttpClientSettings

    async def on_bind(self, event: FactoryBindEvent):
        event.container.add_singleton(AsyncClient, AsyncHttpClient)



from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage
from harp_apps.sqlalchemy_storage.utils.testing.mixins import SqlalchemyStorageTestFixtureMixin


class TestStorageUsage(SqlalchemyStorageTestFixtureMixin):
    async def test_get_usage(self, storage: SqlAlchemyStorage):
        usage = await storage.get_usage()
        assert usage == 0

        for i in range(3):
            await self.create_transaction(storage)

        usage = await storage.get_usage()
        assert usage == 3



import asyncio

from harp.config import Application
from harp.config.events import FactoryBindEvent, FactoryBoundEvent
from harp.typing import Storage
from harp_apps.janitor.worker import JanitorWorker


class JanitorApplication(Application):
    async def on_bind(self, event: FactoryBindEvent):
        pass

    async def on_bound(self, event: FactoryBoundEvent):
        self.worker = JanitorWorker(event.provider.get(Storage))
        self.worker_task = asyncio.create_task(self.worker.run())

    async def unmount(self):
        self.worker.stop()



import os
import subprocess
import threading
from string import Template
from tempfile import NamedTemporaryFile

import httpx
import pytest

from harp import get_logger
from harp.commandline.start import assert_development_packages_are_available
from harp.utils.network import get_available_network_port, wait_for_port
from harp_apps.sqlalchemy_storage.utils.testing.mixins import get_scoped_database_url

logger = get_logger(__name__)


class RunHarpProxyInSubprocessThread(threading.Thread):
    daemon = False

    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, *, daemon=None, config=None):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)

        self.config_filename = None
        if config:
            with NamedTemporaryFile("w+", suffix=".yaml", delete=False) as _tmpfile_config:
                _tmpfile_config.write(config or "")
                self.config_filename = _tmpfile_config.name

        # XXX we may not need dev environment (yet maybe we want to test it works too). To avoid cryptic error, we
        # double check here to get an exception if not available.
        assert_development_packages_are_available()

    def run(self):
        self.process = subprocess.Popen(
            [
                "harp",
                "start",
                "server",
                *(("--file", self.config_filename) if self.config_filename else ()),
                "--disable",
                "telemetry",
                "--disable",
                "dashboard",
            ]
        )

    def join(self, timeout=None):
        # try to kill gracefully ...
        self.process.terminate()
        self.process.wait(5.0)

        # ... and if it does not work, kill it with fire
        self.process.kill()

        # remove temporary config file ...
        if self.config_filename:
            os.unlink(self.config_filename)

        # ... and let threading handle the rest.
        return super().join(timeout)


class AbstractProxyBenchmark:
    config = Template("")

    @pytest.fixture
    async def proxy(self, httpbin, database_url, test_id):
        async with get_scoped_database_url(database_url, test_id) as scoped_database_url:
            port = get_available_network_port()

            try:
                thread = RunHarpProxyInSubprocessThread(
                    config=self.config.substitute(port=port, httpbin=httpbin, database=scoped_database_url)
                )
            except Exception as exc:
                pytest.fail(f"Failed to create subprocess thread: {exc}")

            try:
                try:
                    from pytest_cov.embed import cleanup_on_sigterm
                except ImportError:
                    pass
                else:
                    cleanup_on_sigterm()
                thread.start()
                wait_for_port(port)
                yield f"localhost:{port}"
            finally:
                thread.join()

    def test_noproxy_get(self, benchmark, httpbin):
        @benchmark
        def result():
            return httpx.get(f"{httpbin}/get")

    def test_httpbin_get(self, benchmark, proxy):
        @benchmark
        def result():
            return httpx.get(f"http://{proxy}/get")



"""
Proxy Application

"""

from httpx import AsyncClient

from harp.config.application import Application
from harp.config.events import FactoryBoundEvent

from .controllers import HttpProxyController
from .settings import ProxySettings


class ProxyApplication(Application):
    depends_on = ["harp.services.http"]

    settings_namespace = "proxy"
    settings_type = ProxySettings
    settings: ProxySettings

    @classmethod
    def defaults(cls, settings=None) -> dict:
        settings = settings or super().defaults()
        settings.setdefault("endpoints", [])
        return settings

    async def on_bound(self, event: FactoryBoundEvent):
        for endpoint in self.settings.endpoints:
            event.resolver.add(
                endpoint.port,
                HttpProxyController(
                    endpoint.url,
                    name=endpoint.name,
                    dispatcher=event.dispatcher,
                    http_client=event.provider.get(AsyncClient),
                ),
            )



from harp import Config
from harp.config.application import Application
from harp.utils.identifiers import is_valid_dotted_identifier


def test_is_valid_extension_name():
    # empty
    assert not is_valid_dotted_identifier("")

    # usual
    assert is_valid_dotted_identifier("foo.bar")
    assert is_valid_dotted_identifier("toto")

    # empty part
    assert not is_valid_dotted_identifier(".foo")
    assert not is_valid_dotted_identifier("foo..bar")
    assert not is_valid_dotted_identifier("bar.")

    # dashed/underscored
    assert is_valid_dotted_identifier("o_o")
    assert not is_valid_dotted_identifier("o-o")
    assert is_valid_dotted_identifier("_mazette")
    assert is_valid_dotted_identifier("raise_")


def test_add_application():
    config = Config()
    config.add_application("harp_apps.sqlalchemy_storage")

    assert config.settings == {
        "applications": ["harp_apps.sqlalchemy_storage"],
    }

    # can we serialize it?
    serialized = config.serialize()

    assert serialized == (
        b'{"applications":["harp_apps.sqlalchemy_storage"],"storage":{"type":"sqlalche'
        b'my","url":"sqlite+aiosqlite:///:memory:","migrate":true}}'
    )

    # is the unserialized result the same as before serialization?
    new_config = Config.deserialize(serialized)
    assert new_config == config

    new_config.add_application("foo.bar")

    # Try adding twice
    new_config.add_application("foo.bar")
    assert new_config != config
    assert new_config.settings == {
        "applications": ["harp_apps.sqlalchemy_storage", "foo.bar"],
        "storage": {"migrate": True, "type": "sqlalchemy", "url": "sqlite+aiosqlite:///:memory:"},
    }

    new_config._application_types["foo.bar"] = type("MockedExtension", (Application,), {"name": "foo.bar"})

    assert new_config.serialize() == (
        b'{"applications":["harp_apps.sqlalchemy_storage","foo.bar"],"storage":{"type"'
        b':"sqlalchemy","url":"sqlite+aiosqlite:///:memory:","migrate":tr'
        b"ue}}"
    )



import pytest

from harp.config import DisabledSettings


def test_disabled_settings_cannot_be_enabled():
    disabled_settings = DisabledSettings()
    assert disabled_settings.enabled is False
    with pytest.raises(ValueError):
        disabled_settings.enabled = True



from datetime import UTC, datetime, timedelta

from harp_apps.janitor.settings import OLD_AFTER
from harp_apps.janitor.worker import JanitorWorker
from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage
from harp_apps.sqlalchemy_storage.utils.testing.mixins import SqlalchemyStorageTestFixtureMixin


class TestJanitorWorker(SqlalchemyStorageTestFixtureMixin):
    async def test_delete_old_transactions(self, storage: SqlAlchemyStorage):
        worker = JanitorWorker(storage)

        await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(hours=1))
        await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(minutes=1))
        await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER + timedelta(minutes=1))

        async with storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 3

        await worker.delete_old_transactions()

        async with storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 1

    async def test_delete_orphan_blobs(self, storage: SqlAlchemyStorage):
        worker = JanitorWorker(storage)

        b1 = await self.create_blob(storage, "foo")
        b2 = await self.create_blob(storage, "bar")
        await self.create_blob(storage, "baz")

        async with storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 3
            assert metrics["storage.blobs.orphans"] == 3

        t = await self.create_transaction(storage)
        await self.create_message(storage, transaction_id=t.id, kind="misc", summary="foo", headers=b1.id, body=b2.id)

        async with storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 3
            assert metrics["storage.blobs.orphans"] == 1

        await worker.delete_orphan_blobs()

        async with storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 2
            assert metrics["storage.blobs.orphans"] == 0

    async def test_delete_old_transactions_but_keep_flagged_ones(self, storage: SqlAlchemyStorage):
        worker = JanitorWorker(storage)

        t1 = await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(hours=1))
        await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(minutes=1))
        await self.create_transaction(storage, started_at=datetime.now(UTC) - OLD_AFTER + timedelta(minutes=1))

        user = await storage.users.find_one_by_username("anonymous")
        await storage.flags.create({"type": 1, "user_id": user.id, "transaction_id": t1.id})

        async with storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 3

        await worker.delete_old_transactions()

        async with storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 2



from datetime import UTC
from typing import TYPE_CHECKING

from sqlalchemy import TIMESTAMP, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from harp.http import get_serializer_for
from harp.models.messages import Message as MessageModel

from .base import Base, Repository, with_session

if TYPE_CHECKING:
    from .transactions import Transaction


class Message(Base):
    __tablename__ = "messages"

    id = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)
    kind = mapped_column(String(10))
    summary = mapped_column(String(255))
    headers = mapped_column(String(40))
    body = mapped_column(String(40))
    created_at = mapped_column(TIMESTAMP(timezone=True))

    transaction_id = mapped_column(ForeignKey("transactions.id", ondelete="CASCADE"))
    transaction: Mapped["Transaction"] = relationship(back_populates="messages")

    def to_model(self):
        return MessageModel(
            id=self.id,
            transaction_id=self.transaction_id,
            kind=self.kind,
            summary=self.summary,
            headers=self.headers,
            body=self.body,
            created_at=self.created_at.replace(tzinfo=UTC) if self.created_at else self.created_at,
        )

    @classmethod
    def from_models(cls, transaction, message, headers, content):
        serializer = get_serializer_for(message)

        obj = cls()
        obj.transaction_id = transaction.id
        obj.kind = message.kind
        obj.summary = serializer.summary
        obj.headers = headers.id
        obj.body = content.id
        obj.created_at = message.created_at.astimezone(UTC)
        return obj


class MessagesRepository(Repository[Message]):
    Type = Message

    @with_session
    async def create(self, values: dict | MessageModel, /, *, session):
        if isinstance(values, MessageModel):
            values = dict(
                id=values.id,
                transaction_id=values.transaction_id,
                kind=values.kind,
                summary=values.summary,
                headers=values.headers,
                body=values.body,
                created_at=values.created_at,
            )
        return await super().create(values, session=session)



from _pytest.fixtures import fixture

from harp import Config


class BaseTestForApplications:
    name = None
    config_key = None

    expected_defaults = {}

    @fixture
    def configuration(self):
        config = Config()
        config.add_application(self.name)
        return config

    @fixture
    def ApplicationType(self, configuration):
        return configuration.get_application_type(self.name)

    def test_default_settings_on_none_passed(self, ApplicationType):
        assert ApplicationType.defaults() == self.expected_defaults

    def test_default_settings_on_empty_dict_passed(self, ApplicationType):
        settings = {}
        assert ApplicationType.defaults(settings) == self.expected_defaults

    def test_instanciate_with_default_settings(self, ApplicationType):
        defaults = ApplicationType.defaults()
        application = ApplicationType(defaults)
        assert application.validate() == defaults



from httpx import AsyncClient

from harp_apps.http_client.settings import HttpClientSettings


class AsyncHttpClient(AsyncClient):
    settings: HttpClientSettings

    def __init__(self, settings: HttpClientSettings):
        transport = settings.transport.build()

        if settings.cache.enabled:
            transport = settings.cache.transport.build(
                transport=transport,
                controller=settings.cache.controller.build(),
                storage=settings.cache.storage.build(),
            )

        super().__init__(transport=transport, timeout=settings.timeout)



import pytest

from harp.errors import ConfigurationError

from ..settings import DashboardAuthSetting


def test_no_auth():
    assert DashboardAuthSetting() is None
    assert DashboardAuthSetting(type="") is None
    assert DashboardAuthSetting(type=None) is None


def test_invalid_auth():
    with pytest.raises(ConfigurationError):
        DashboardAuthSetting(type=None, value="no chance")

    with pytest.raises(ConfigurationError):
        DashboardAuthSetting(type="invalid")


def test_basic_auth():
    assert DashboardAuthSetting(
        type="basic",
        algorithm="plain",
        users={"foo": "bar"},
    ).to_dict() == {
        "type": "basic",
        "algorithm": "plain",
        "users": {"foo": "bar"},
    }



import orjson
from multidict import CIMultiDict

from harp.utils.bytes import ensure_bytes

from .typing import BaseHttpMessage


class HttpResponse(BaseHttpMessage):
    kind = "response"

    def __init__(self, body: bytes | str, /, *, status: int = 200, headers: dict = None, content_type=None):
        super().__init__()

        self._body = ensure_bytes(body)
        self._status = int(status)
        self._headers = CIMultiDict(headers or {})

        if content_type:
            self._headers["content-type"] = content_type

    @property
    def body(self) -> bytes:
        return self._body

    @property
    def status(self) -> int:
        return self._status

    @property
    def headers(self) -> CIMultiDict:
        return self._headers

    @property
    def content_type(self) -> str:
        return self._headers.get("content-type", "text/plain")


class JsonHttpResponse(HttpResponse):
    def __init__(self, body: dict, /, *, status: int = 200, headers: dict = None):
        super().__init__(orjson.dumps(body), status=status, headers=headers, content_type="application/json")


class AlreadyHandledHttpResponse(HttpResponse):
    def __init__(self):
        super().__init__(b"")



import sys

from harp.utils.processes import check_output


def parse_dependencies(dependencies: list[str]) -> dict[str, str]:
    """Transform a list of dependencies into a dictionary of dependencies, with package name as key."""
    parsed = {}

    for dependency in dependencies:
        if dependency.startswith("-e"):
            pkg_repo, pkg_name = dependency.rsplit("/", 1)
            if pkg_name in parsed:
                raise ValueError(f"Duplicate package name {pkg_name} in dependencies.")
            parsed[pkg_name] = pkg_repo + "/" + pkg_name
        else:
            pkg_name, pkg_version = dependency.split("==")
            if pkg_name in parsed:
                raise ValueError(f"Duplicate package name {pkg_name} in dependencies.")
            parsed[pkg_name] = pkg_version
    return parsed


async def get_python_dependencies():
    return list(
        filter(
            lambda x: x and not x.startswith("#"),
            (await check_output(sys.executable, "-m", "pip", "freeze")).decode("utf-8").split("\n"),
        ),
    )



from unittest.mock import ANY, AsyncMock, patch

import pytest
import respx
from httpx import AsyncClient, Response
from whistle import AsyncEventDispatcher

from harp.utils.bytes import ensure_bytes
from harp.utils.testing.mixins import ControllerTestFixtureMixin
from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage
from harp_apps.sqlalchemy_storage.utils.testing.mixins import SqlalchemyStorageTestFixtureMixin

from ..controllers import HttpProxyController
from ..events import EVENT_TRANSACTION_STARTED


class DispatcherTestFixtureMixin:
    @pytest.fixture(scope="function")
    def dispatcher(self):
        return AsyncEventDispatcher()


class HttpProxyControllerTestFixtureMixin(ControllerTestFixtureMixin):
    ControllerType = HttpProxyController

    @pytest.fixture(autouse=True)
    def setup(self):
        # forces the user agent to be a known value, even if versions are incremented
        with patch("harp_apps.proxy.controllers.HttpProxyController.user_agent", "test/1.0"):
            yield

    def mock_http_endpoint(self, database_url, /, *, status=200, content=""):
        """Make sure you decorate your tests function using this with respx.mock decorator, otherwise the real network
        will be called and you may have some headaches..."""
        return respx.get(database_url).mock(return_value=Response(status, content=ensure_bytes(content)))

    def create_controller(self, url=None, *args, http_client=None, **kwargs):
        return super().create_controller(
            url or "http://example.com/", *args, http_client=http_client or AsyncClient(), **kwargs
        )


class TestHttpProxyController(HttpProxyControllerTestFixtureMixin, DispatcherTestFixtureMixin):
    @respx.mock
    async def test_basic_get(self):
        endpoint = self.mock_http_endpoint("http://example.com/", content="Hello.")
        request, response = await self.call_controller(self.create_controller("http://example.com/"))

        # check output and side effects
        assert endpoint.called and endpoint.call_count == 1
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b"Hello."

    @respx.mock
    async def test_get_with_tags(self, dispatcher: AsyncEventDispatcher):
        endpoint = self.mock_http_endpoint("http://example.com/", content="Hello.")

        # register a mock handler to inspect the actually created transaction
        transaction_started_handler = AsyncMock()
        dispatcher.add_listener(EVENT_TRANSACTION_STARTED, transaction_started_handler)

        # call our controller
        request, response = await self.call_controller(
            self.create_controller("http://example.com/", dispatcher=dispatcher),
            headers={
                "x-harp-foo": "bar",
                "accept": "application/json",
                "vary": "custom",
            },
        )

        # check that our remote endpoint was passed custom headers, but not internal ones
        assert endpoint.called and endpoint.call_count == 1
        assert "x-harp-foo" not in endpoint.calls[0].request.headers
        assert endpoint.calls[0].request.headers["accept"] == "application/json"
        assert endpoint.calls[0].request.headers["vary"] == "custom"

        # check that the transaction was tagged with the expected values
        assert transaction_started_handler.called and transaction_started_handler.call_count == 1
        assert transaction_started_handler.call_args.args[0].transaction.tags == {"foo": "bar"}

        # check we got a valid response
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b"Hello."


class TestHttpProxyControllerWithStorage(
    HttpProxyControllerTestFixtureMixin,
    SqlalchemyStorageTestFixtureMixin,
    DispatcherTestFixtureMixin,
):
    async def _find_one_transaction_with_messages_from_storage(self, storage):
        # get transaction, request and response
        transactions = await storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        transaction = transactions[0]
        return transaction, transaction.messages[0], transaction.messages[1]

    @respx.mock
    async def test_basic_get(self, storage: SqlAlchemyStorage):
        self.mock_http_endpoint("http://example.com/", content="Hello.")

        # register the storage
        await self.call_controller(self.create_controller("http://example.com/", dispatcher=storage._dispatcher))
        await storage.wait_for_background_tasks_to_be_processed()

        transaction, request, response = await self._find_one_transaction_with_messages_from_storage(storage)

        # transaction
        assert transaction.to_dict(omit_none=False) == {
            "id": ANY,
            "type": "http",
            "endpoint": None,
            "elapsed": ANY,
            "apdex": ANY,
            "started_at": ANY,
            "finished_at": ANY,
            "messages": ANY,
            "tags": {},
            "extras": {"flags": [], "method": "GET", "status_class": "2xx", "cached": False, "no_cache": False},
        }

        # request
        assert (await storage.get_blob(request.headers)).data == b"host: example.com\nuser-agent: test/1.0"
        assert (await storage.get_blob(request.body)).data == b""
        assert request.to_dict(omit_none=False) == {
            "id": 1,
            "transaction_id": ANY,
            "kind": "request",
            "summary": "GET / HTTP/1.1",
            "headers": "ecd46dd3023926417e1205b219aec4c2b9e0dab0",
            "body": "adc83b19e793491b1c6ea0fd8b46cd9f32e592fc",
            "created_at": ANY,
        }

        # response
        assert (await storage.get_blob(response.headers)).data == b""
        assert (await storage.get_blob(response.body)).data == b"Hello."
        assert response.to_dict(omit_none=False) == {
            "id": 2,
            "transaction_id": ANY,
            "kind": "response",
            "summary": "HTTP/1.1 200 OK",
            "headers": "7507b0bca78329128f61f37d09c88b8712cecd64",
            "body": "6ffdd89703735cc316470566467b816446f008ce",
            "created_at": ANY,
        }

    @respx.mock
    async def test_get_with_tags(self, storage):
        self.mock_http_endpoint("http://example.com/", content="Hello.")

        # call our controller
        await self.call_controller(
            self.create_controller("http://example.com/", dispatcher=storage._dispatcher),
            headers={
                "x-harp-foo": "bar",
                "accept": "application/json",
                "vary": "custom",
            },
        )
        await storage.wait_for_background_tasks_to_be_processed()

        transaction, request, response = await self._find_one_transaction_with_messages_from_storage(storage)

        assert transaction.to_dict(omit_none=False) == {
            "id": ANY,
            "type": "http",
            "endpoint": None,
            "elapsed": ANY,
            "apdex": ANY,
            "started_at": ANY,
            "finished_at": ANY,
            "messages": ANY,
            "tags": {"foo": "bar"},
            "extras": {"flags": [], "method": "GET", "status_class": "2xx", "cached": False, "no_cache": False},
        }

        assert (await storage.get_blob(request.headers)).data == (
            b"accept: application/json\nvary: custom\nhost: example.com\nuser-agent: test/1.0"
        )
        assert (await storage.get_blob(request.body)).data == b""
        assert request.to_dict(omit_none=False) == {
            "id": 1,
            "transaction_id": ANY,
            "kind": "request",
            "summary": "GET / HTTP/1.1",
            "headers": "885ad9e4633a3c038f13c32e4ee48fc9221c0152",
            "body": "adc83b19e793491b1c6ea0fd8b46cd9f32e592fc",
            "created_at": ANY,
        }

        assert (await storage.get_blob(response.headers)).data == b""
        assert (await storage.get_blob(response.body)).data == b"Hello."
        assert response.to_dict(omit_none=False) == {
            "id": 2,
            "transaction_id": ANY,
            "kind": "response",
            "summary": "HTTP/1.1 200 OK",
            "headers": "7507b0bca78329128f61f37d09c88b8712cecd64",
            "body": "6ffdd89703735cc316470566467b816446f008ce",
            "created_at": ANY,
        }



from dataclasses import dataclass

from harp.utils.env import cast_bool


@dataclass
class SqlAlchemyStorageSettings:
    type: str = "sqlalchemy"
    url: str = "sqlite+aiosqlite:///harp.db"
    migrate: bool = True

    def __post_init__(self):
        self.migrate = cast_bool(self.migrate)



import dataclasses
import socket
import time


def get_available_network_port():
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def wait_for_port(port: int, host: str = "localhost", timeout: float = 10.0):
    """Wait until a port starts accepting TCP connections.
    Args:
        port: Port number.
        host: Host address on which the port should exist.
        timeout: In seconds. How long to wait before raising errors.
    Raises:
        TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError(
                    "Waited too long for the port {} on host {} to start accepting " "connections.".format(port, host)
                ) from ex


@dataclasses.dataclass(frozen=True)
class Bind:
    host: str
    port: int

    def __str__(self):
        return f"{self.host}:{self.port}"



"""
The Command Line (:mod:`harp.commandline`) package bundles everything related to the `harp` command.

It contains an :func:`entrypoint` callable, built using `click`, which is called when you type `harp` in your terminal.

It is a high-level development helper: for example, `harp start` wraps `honcho`, a python process manager (foreman
clone) that can spawn multiple processes to help you with development.

Production environment will favour the lower-level `bin/entrypoint` script, that skips the process manager entirely
and runs the application server directly.

This may be subject to changes in the future (especially to align/refactor arg parsers), but this approach works
well for now.

.. todo::

    @click.option('--prod/--dev', default=False, help="Set defaults for production or development environment (default
    to dev).")

    @click.option('--verbose', '-v', count=True)
    --verbose : change default log verbosity

    @click.option('--debug', is_flag=True)
    --debug : ?

    Manage process list / options using a class (with hierarchy for prod/dev, for example)
    If one process only, no honcho ? Or mayne no honcho manager ?

.. todo::

    Possible future things (or not)

    - build (compile stuff for production)
    - serve (run a production-like server)

    Document how harp start work (honcho, devserver port, ...)

"""

from harp.commandline.server import server
from harp.settings import HARP_ENV
from harp.utils.commandline import check_packages, click

__title__ = "Command Line"

IS_DEVELOPMENT_ENVIRONMENT = False

if HARP_ENV == "dev" or check_packages("honcho", "watchfiles"):
    IS_DEVELOPMENT_ENVIRONMENT = True

if HARP_ENV == "prod":
    IS_DEVELOPMENT_ENVIRONMENT = False


@click.group()
def entrypoint():
    """HTTP Application Runtime Proxy (HARP)

    The following commands are available to help you setup and run your HTTP proxy application.

    """
    pass


if IS_DEVELOPMENT_ENVIRONMENT:
    from harp.commandline.start import start

    entrypoint.add_command(start)

    from harp.commandline.install import install_dev

    entrypoint.add_command(install_dev)

if check_packages("alembic"):
    from harp.commandline.migrations import create_migration, feature, history, migrate

    entrypoint.add_command(migrate)
    entrypoint.add_command(feature)
    entrypoint.add_command(history)

    if IS_DEVELOPMENT_ENVIRONMENT:
        entrypoint.add_command(create_migration)


entrypoint.add_command(server)

__all__ = [
    "entrypoint",
]



from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

from whistle import Event

from harp.http import BaseMessage, HttpRequest
from harp.models.transactions import Transaction

if TYPE_CHECKING:
    from harp.http import HttpResponse

EVENT_CORE_STARTED = "core.started"


class RequestEvent(Event):
    name = "core.request"

    def __init__(self, request: HttpRequest):
        self._request = request
        self._controller = None

    @cached_property
    def request(self):
        return self._request

    @property
    def controller(self):
        return self._controller

    def set_controller(self, controller):
        self._controller = controller


EVENT_CORE_REQUEST = RequestEvent.name


class ControllerEvent(RequestEvent):
    name = "core.controller"

    def __init__(self, request: HttpRequest, controller):
        super().__init__(request)
        self._controller = controller


EVENT_CORE_CONTROLLER = ControllerEvent.name


class ResponseEvent(RequestEvent):
    name = "core.response"

    def __init__(self, request: HttpRequest, response):
        super().__init__(request)
        self.response = response


EVENT_CORE_RESPONSE = ResponseEvent.name


class ControllerViewEvent(RequestEvent):
    """
    The view event allows to transform controller return values into response objects.
    """

    name = "controller.view"

    value: Any
    response: Optional["HttpResponse"]

    def __init__(self, request: HttpRequest, value: Any):
        super().__init__(request)

        self.value = value
        self.response = None

    def set_response(self, response: "HttpResponse"):
        self.response = response
        self.stop_propagation()


EVENT_CONTROLLER_VIEW = ControllerViewEvent.name


class TransactionEvent(Event):
    def __init__(self, transaction: Transaction):
        self.transaction = transaction


class MessageEvent(TransactionEvent):
    def __init__(self, transaction: Transaction, message: BaseMessage):
        super().__init__(transaction)
        self.message = message



import asyncio
import re
from contextlib import asynccontextmanager
from datetime import UTC, datetime, timedelta
from functools import partial
from operator import itemgetter
from typing import Iterable, List, Optional, TypedDict, override

from sqlalchemy import and_, bindparam, case, delete, func, literal, literal_column, null, or_, select, text, update
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine
from sqlalchemy.sql.functions import count
from whistle import IAsyncEventDispatcher

from harp import get_logger
from harp.asgi.events import EVENT_CORE_STARTED, MessageEvent, TransactionEvent
from harp.http import get_serializer_for
from harp.models import Blob as BlobModel
from harp.models.base import Results
from harp.models.transactions import Transaction as TransactionModel
from harp.settings import PAGE_SIZE
from harp.typing.storage import Storage
from harp.utils.background import AsyncWorkerQueue
from harp.utils.dates import ensure_datetime
from harp.utils.tpdex import tpdex
from harp_apps.proxy.events import EVENT_TRANSACTION_ENDED, EVENT_TRANSACTION_MESSAGE, EVENT_TRANSACTION_STARTED

from .constants import TimeBucket
from .models import (
    FLAGS_BY_NAME,
    Base,
    Blob,
    BlobsRepository,
    FlagsRepository,
    Message,
    MessagesRepository,
    MetricsRepository,
    MetricValuesRepository,
    TagsRepository,
    TagValuesRepository,
    Transaction,
    TransactionsRepository,
    User,
    UserFlag,
    UsersRepository,
)
from .settings import SqlAlchemyStorageSettings
from .utils.dates import TruncDatetime


class TransactionsGroupedByTimeBucket(TypedDict):
    datetime: datetime | None
    count: int
    errors: int
    meanDuration: float


logger = get_logger(__name__)

_FILTER_COLUMN_NAMES = {
    "method": "x_method",
    "status": "x_status_class",
}


def _filter_query(query, name, values):
    if values:
        query = query.filter(
            getattr(
                Transaction,
                _FILTER_COLUMN_NAMES.get(name, name),
            ).in_(values)
        )
    return query


def _filter_query_for_user_flags(query, values, /, *, user_id):
    if values:
        if "NULL" in values:
            query = query.outerjoin(UserFlag).filter(
                or_(
                    and_(
                        UserFlag.user_id == user_id,
                        UserFlag.type.in_(
                            list(
                                map(FLAGS_BY_NAME.get, values),
                            )
                        ),
                    ),
                    or_(UserFlag.user_id != user_id, (UserFlag.type.is_(null()))),
                )
            )
        else:
            query = query.join(UserFlag).filter(
                UserFlag.user_id == user_id,
                UserFlag.type.in_(
                    list(
                        map(FLAGS_BY_NAME.get, values),
                    )
                ),
            )
    return query


def _filter_transactions_based_on_text(query, search_text: str, dialect_name: str):
    # Escape special characters in search_text
    search_text = re.sub(r"([-\*\(\)~\"@<>\^+]+)", r"", search_text)
    query = query.join(Message, isouter=True)
    # check dialect and use appropriate full text search
    if dialect_name == "mysql":
        return query.filter(
            text(
                f"MATCH ({Transaction.__tablename__}.endpoint) "
                f"AGAINST (:search_text IN BOOLEAN MODE) OR "
                f"MATCH ({Message.__tablename__}.summary) "
                f"AGAINST (:search_text IN BOOLEAN MODE)",
            ).bindparams(bindparam("search_text", literal_column(f"'{search_text}*'")))
        )

    return query.filter(
        (Transaction.endpoint.ilike(bindparam("search_text", f"%{search_text}%")))
        | Message.summary.ilike(bindparam("search_text", f"%{search_text}%"))
    )


class SqlAlchemyStorage(Storage):
    """
    Storage implementation using SQL Alchemy Core, with async drivers.

    Currently supported/tested database drivers:

    - aiosqlite (sqlite+aiosqlite://...)

    """

    engine: AsyncEngine
    """Reference to the sqlalchemy async engine, which is a gateway to the database connectivity, able to provide a
    connection used to execute queries."""

    def __init__(self, dispatcher: IAsyncEventDispatcher, settings: SqlAlchemyStorageSettings):
        self.settings = settings
        self.metadata = Base.metadata

        self.engine = create_async_engine(self.settings.url)
        self.session_factory = async_sessionmaker(self.engine, expire_on_commit=False)

        # TODO is this the right place ? (maybe it is, but maybe it causes tight coupling to the ed which may be not
        #      right, especially if we want to be able to use the storage out of the special context of harp proxy)
        dispatcher.add_listener(EVENT_CORE_STARTED, self._on_startup_actions, priority=-20)
        dispatcher.add_listener(EVENT_TRANSACTION_STARTED, self._on_transaction_started)
        dispatcher.add_listener(EVENT_TRANSACTION_MESSAGE, self._on_transaction_message)
        dispatcher.add_listener(EVENT_TRANSACTION_ENDED, self._on_transaction_ended)
        self._dispatcher = dispatcher

        self._is_ready = asyncio.Event()
        self._worker = None

        self.blobs = BlobsRepository(self.session_factory)
        self.messages = MessagesRepository(self.session_factory)
        self.tags = TagsRepository(self.session_factory)
        self.tag_values = TagValuesRepository(self.session_factory)
        self.transactions = TransactionsRepository(self.session_factory, tags=self.tags, tag_values=self.tag_values)
        self.users = UsersRepository(self.session_factory)
        self.metrics = MetricsRepository(self.session_factory)
        self.metric_values = MetricValuesRepository(self.session_factory)
        self.flags = FlagsRepository(self.session_factory)

        self._debug = False

    @asynccontextmanager
    async def begin(self):
        async with self.session_factory() as session:
            async with session.begin():
                yield session

    def install_debugging_instrumentation(self, *, echo=False):
        self._debug = True
        self._original_session_factory = self.session_factory
        self._debug_index = 1
        self.sql_queries = []

        if echo:
            from rich.console import Console
            from rich.syntax import Syntax

            _console = Console(force_terminal=True, width=180)

        @asynccontextmanager
        async def _session_factory():
            async with self._original_session_factory() as session:
                original_execute = session.sync_session._execute_internal

                def _instrumented_execute(statement, *args, **kwargs):
                    sql_query = str(
                        statement.compile(session.sync_session.bind, compile_kwargs={"literal_binds": True})
                    )
                    self.sql_queries.append(sql_query)

                    final_result = original_execute(statement, *args, **kwargs)

                    if echo:
                        _console.print(f"🛢 SQL QUERY (#{self._debug_index})", style="bold")
                        _console.print(
                            Syntax(
                                sql_query,
                                "sql",
                                word_wrap=True,
                                theme="vs",
                            )
                        )

                    if echo and self.engine.dialect.name == "postgresql":
                        _get0 = itemgetter(0)
                        _console.print(f"🛢 EXPLAIN ANALYZE (#{self._debug_index})", style="bold")
                        _console.print(
                            Syntax(
                                "\n".join(
                                    map(_get0, original_execute(text("EXPLAIN ANALYZE " + sql_query)).fetchall())
                                ),
                                "sql",
                                word_wrap=True,
                                theme="vs",
                            )
                        )

                    self._debug_index += 1
                    return final_result

                session.sync_session._execute_internal = _instrumented_execute
                try:
                    yield session
                finally:
                    session.sync_session._execute_internal = original_execute

        self.session_factory = _session_factory

    async def initialize(self):
        """Initialize database."""
        if self.settings.migrate:
            await self._run_migrations()
        await self.create_users(["anonymous"])
        self._is_ready.set()

    async def _run_migrations(self):
        """Convenience helper to run the migrations. This behaviour can be disabled by setting migrate=false in the
        storage settings."""
        from alembic import command

        from harp_apps.sqlalchemy_storage.utils.migrations import create_alembic_config

        from .utils.migrations import do_migrate

        alembic_cfg = create_alembic_config(self.engine.url.render_as_string(hide_password=False))

        migrator = partial(command.upgrade, alembic_cfg, "head")

        await do_migrate(self.engine, migrator=migrator)

    @property
    def worker(self):
        if not self._worker:
            self._worker = AsyncWorkerQueue()
        return self._worker

    async def wait_for_background_tasks_to_be_processed(self):
        if self._worker:
            await self._worker.wait_until_empty()

    @override
    async def get_facet_meta(self, name):
        if name == "endpoint":
            # get transaction count grouped by endpoint
            async with self.begin() as session:
                query = select(Transaction.endpoint, func.count()).group_by(Transaction.endpoint)
                result = await session.execute(query)
                return {row[0]: row[1] for row in result.fetchall()}

        raise NotImplementedError(f"Unknown facet: {name}")

    @override
    async def get_transaction_list(
        self,
        *,
        username: str,
        with_messages=False,
        filters=None,
        page: int = 1,
        cursor: str = "",
        text_search="",
    ):
        """
        Implements :meth:`Storage.find_transactions <harp.typing.storage.Storage.find_transactions>`.

        """

        user = await self.users.find_one_by_username(username)

        result = Results()
        query = self.transactions.select(
            with_messages=with_messages,
            with_user_flags=user.id if user else False,
            with_tags=True,
        )

        if filters:
            query = _filter_query(query, "endpoint", filters.get("endpoint", None))
            query = _filter_query(query, "method", filters.get("method", None))
            query = _filter_query(query, "status", filters.get("status", None))
            query = _filter_query_for_user_flags(query, filters.get("flag", None), user_id=user.id)

        if text_search:
            query = _filter_transactions_based_on_text(query, text_search, dialect_name=self.engine.dialect.name)

        query = query.order_by(Transaction.started_at.desc())

        # apply cursor (before count)
        if page and cursor:
            query = query.filter(Transaction.id <= cursor)

        async with self.begin() as session:
            # count items from query
            result.meta["total"] = await session.scalar(
                query.with_only_columns(func.count(Transaction.id)).order_by(None)
            )

        # apply limit/offset (after count)
        query = query.limit(PAGE_SIZE)
        if page:
            query = query.offset(max(0, (page - 1) * PAGE_SIZE))

        async with self.begin() as session:
            for transaction in (await session.scalars(query)).unique().all():
                result.append(transaction.to_model(with_user_flags=True))

        return result

    @override
    async def get_transaction(self, id: str, /, *, username: str) -> Optional[TransactionModel]:
        user = await self.users.find_one_by_username(username)

        return (
            await self.transactions.find_one_by_id(
                id,
                with_messages=True,
                with_user_flags=user.id if user else False,
                with_tags=True,
            )
        ).to_model(with_user_flags=True)

    @override
    async def transactions_grouped_by_time_bucket(
        self,
        endpoint: Optional[str] = None,
        time_bucket: str = TimeBucket.DAY.value,
        start_datetime: Optional[datetime] = None,
    ) -> List[TransactionsGroupedByTimeBucket]:
        if time_bucket not in [e.value for e in TimeBucket]:
            raise ValueError(
                f"Invalid time bucket: {time_bucket}. Must be one of {', '.join([e.value for e in TimeBucket])}."
            )

        s_date = TruncDatetime(literal(time_bucket), Transaction.started_at).label("tb")
        query = select(
            s_date,
            func.count(),
            func.sum(case((Transaction.x_status_class.in_(("5xx", "ERR")), 1), else_=0)),
            func.avg(Transaction.elapsed),
            func.avg(Transaction.apdex),
        )

        if endpoint:
            query = query.where(Transaction.endpoint == endpoint)

        if start_datetime:
            query = query.where(Transaction.started_at >= start_datetime.astimezone(UTC))

        query = query.group_by(s_date).order_by(s_date.asc())
        async with self.begin() as session:
            result = await session.execute(query)
            return [
                {
                    "datetime": ensure_datetime(row[0], UTC),
                    "count": row[1],
                    "errors": int(row[2]),
                    "meanDuration": row[3] if row[3] else 0,
                    "meanApdex": row[4],
                    # ! probably sqlite struggling with unfinished transactions
                }
                for row in result.fetchall()
            ]

    async def get_usage(self):
        async with self.begin() as session:
            query = select(count(Transaction.id)).where(
                Transaction.started_at > (datetime.now(UTC) - timedelta(hours=24))
            )
            return (await session.execute(query)).scalar_one_or_none()

    @override
    async def get_blob(self, blob_id):
        """
        Retrieve a blob from the database, using its hash.
        Returns None if not found.

        :param blob_id: sha1 hash of the blob
        :return: Blob or None
        """
        async with self.begin() as session:
            row = (
                await session.execute(
                    select(Blob).where(Blob.id == blob_id),
                )
            ).fetchone()

        if row:
            return BlobModel(id=blob_id, data=row[0].data, content_type=row[0].content_type)

    async def _on_startup_actions(self, TransactionEvent):
        """Event handler to create the database tables on startup. May drop them first if configured to do so."""
        await self.initialize()

    @override
    async def set_user_flag(self, *, transaction_id: str, username: str, flag: int, value=True):
        """Sets or unsets a user flag on a transaction."""
        async with self.begin() as session:
            user = await self.users.find_one_by_username(username)
            transaction = await self.transactions.find_one_by_id(transaction_id)

            if value:
                session.add(
                    UserFlag(
                        transaction_id=transaction.id,
                        user_id=user.id,
                        type=flag,
                    )
                )
            else:
                await session.execute(
                    delete(UserFlag).where(UserFlag.transaction_id == transaction.id, UserFlag.user_id == user.id)
                )

    async def _on_transaction_started(self, event: TransactionEvent):
        """Event handler to store the transaction in the database."""
        return await self.transactions.create(event.transaction)

    async def _on_transaction_message(self, event: MessageEvent):
        await event.message.join()
        serializer = get_serializer_for(event.message)

        # todo is the "__headers__" dunder content type any good idea ? maybe it's just a waste of bytes.
        headers_blob = BlobModel.from_data(serializer.headers, content_type="__headers__")
        content_blob = BlobModel.from_data(serializer.body, content_type=event.message.headers.get("content-type"))

        def create_store_blob_task(blob):
            async def store_blob_task():
                async with self.begin() as session:
                    if not (
                        await session.execute(select(select(Blob.id).where(Blob.id == blob.id).exists()))
                    ).scalar_one():
                        db_blob = Blob()
                        db_blob.id = blob.id
                        db_blob.content_type = blob.content_type
                        db_blob.data = blob.data
                        session.add(db_blob)

            return store_blob_task

        await self.worker.push(create_store_blob_task(headers_blob), ignore_errors=False)
        await self.worker.push(create_store_blob_task(content_blob), ignore_errors=True)

        async def store_message_task():
            async with self.begin() as session:
                session.add(
                    Message.from_models(event.transaction, event.message, headers_blob, content_blob),
                )

        await self.worker.push(store_message_task)

    async def _on_transaction_ended(self, event: TransactionEvent):
        async def finalize_transaction():
            async with self.begin() as session:
                await session.execute(
                    update(Transaction)
                    .where(Transaction.id == event.transaction.id)
                    .values(
                        finished_at=event.transaction.finished_at.astimezone(UTC),
                        elapsed=event.transaction.elapsed,
                        apdex=tpdex(event.transaction.elapsed),
                        x_status_class=event.transaction.extras.get("status_class"),
                        x_cached=event.transaction.extras.get("cached"),
                    )
                )

        await self.worker.push(finalize_transaction)

    async def ready(self):
        await self._is_ready.wait()

    @override
    async def create_users_once_ready(self, users: Iterable[str]):
        """Sets the list of users to be created once the database is ready."""

        async def defered_create_users():
            await self.ready()
            await self.create_users(users)

        await self.worker.push(defered_create_users)

    async def create_users(self, users: Iterable[str]):
        async with self.begin() as session:
            for username in users:
                # Check if the username already exists
                result = await session.execute(select(User).where(User.username == username))
                existing_user = result.scalar_one_or_none()

                # If the username does not exist, create a new user
                if existing_user is None:
                    user = User()
                    user.username = username
                    session.add(user)



import pytest

from harp.utils.testing.applications import BaseTestForApplications


class TestSqlAlchemyStorageApplication(BaseTestForApplications):
    name = "harp_apps.sqlalchemy_storage"
    config_key = "storage"

    expected_defaults = {
        "type": "sqlalchemy",
        "url": "sqlite+aiosqlite:///:memory:",
        "migrate": True,
    }

    @pytest.mark.parametrize(
        "settings",
        [
            {"type": "foo"},
            {"something": "different"},
        ],
    )
    def test_defaults_does_not_change_anything_if_not_sqlalchemy_type(self, ApplicationType, settings):
        assert ApplicationType.defaults(settings) == settings

    @pytest.mark.parametrize(
        "settings",
        [
            {"type": "sqlalchemy", "url": "sqlite+aiosqlite:///harp.db"},
            {"type": "sqlalchemy", "migrate": True},
            {"type": "sqlalchemy", "migrate": False},
        ],
    )
    def test_defaults_fills_missing_values_for_sqlalchemy_type(self, ApplicationType, settings):
        assert ApplicationType.defaults(settings) == {**self.expected_defaults, **settings}



from dataclasses import dataclass
from decimal import Decimal

import pytest
from hishel import Controller

from harp.config import BaseSetting, Definition, Lazy, asdict


class TestFactoryDefinition:
    def test_not_existing(self):
        definition = Lazy("foo:bar")
        assert isinstance(definition, Definition)

        assert definition.path == "foo"
        assert definition.name == "bar"
        assert len(definition.args) == 0
        assert len(definition.kwargs) == 0

        with pytest.raises(ImportError):
            assert definition.validate()

    def test_existing(self):
        definition = Lazy("harp.config:Application", {"foo": "bar"})
        assert isinstance(definition, Definition)

        assert definition.path == "harp.config"
        assert definition.name == "Application"
        assert len(definition.args) == 1
        assert len(definition.kwargs) == 0

        assert definition.validate()

        instance = definition.build()
        assert instance.settings == {"foo": "bar"}

    def test_using_type(self):
        definition = Lazy(Decimal)
        assert isinstance(definition, Definition)

        assert definition.path == "decimal"
        assert definition.name == "Decimal"

        assert definition.build("42.0") == 42

    def test_usage(self):
        @dataclass(kw_only=True)
        class MySetting:
            dep: Definition[Decimal] = Lazy("decimal:Decimal")

            def __post_init__(self):
                self.dep = Lazy(self.dep)

        setting = MySetting()
        assert setting.dep.build() == 0

        class OtherDecimal(Decimal):
            pass

        setting = MySetting(dep=OtherDecimal)
        assert setting.dep.build() == 0

        setting = MySetting(dep="decimal:Decimal")
        assert setting.dep.build() == 0

        setting = MySetting(dep={"@type": "decimal:Decimal"})
        assert setting.dep.build() == 0

        setting = MySetting(dep={"@type": "decimal:Decimal", "@args": ["42.0"]})
        assert setting.dep.build() == 42

    def test_reallife(self):
        @dataclass(kw_only=True)
        class MySetting:
            dep: Definition[Controller] = Lazy("hishel:Controller")

            def __post_init__(self):
                self.dep = Lazy(self.dep)

        setting = MySetting()
        assert setting.dep.build()._force_cache is False

        class MyController(Controller):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._force_cache = True

        setting = MySetting(dep={"@type": MyController})
        assert setting.dep.build()._force_cache is True

    def test_serialize(self):
        @dataclass
        class Foo(BaseSetting):
            dep: Definition = Lazy("decimal:Decimal")

        assert asdict(Foo()) == {
            "dep": {
                "@type": "decimal:Decimal",
            }
        }

        assert asdict(
            Foo(
                dep="io:StringIO",
            )
        ) == {
            "dep": {
                "@type": "io:StringIO",
            }
        }

        assert asdict(
            Foo(
                dep={"@type": "io:StringIO"},
            )
        ) == {
            "dep": {
                "@type": "io:StringIO",
            }
        }

        assert asdict(
            Foo(
                dep={"@type": "math:gcd", "@args": [10, 20, 30]},
            )
        ) == {
            "dep": {
                "@type": "math:gcd",
                "@args": (10, 20, 30),
            }
        }

        assert asdict(
            Foo(
                dep={"@type": "foo:bar", "@args": [1, 2, 3, 5, 8], "algorithm": "fibonacci"},
            )
        ) == {
            "dep": {
                "@args": (1, 2, 3, 5, 8),
                "@type": "foo:bar",
                "algorithm": "fibonacci",
            },
        }



import logging.config
import os
from typing import Any

import structlog

shared_processors = [
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.ExtraAdder(),
    structlog.processors.TimeStamper(fmt="iso"),
]

processors = [
    structlog.stdlib.filter_by_level,
    structlog.contextvars.merge_contextvars,
    *shared_processors,
    structlog.dev.set_exc_info,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.UnicodeDecoder(),
    structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
]

try:
    _exc_formatter_options = {"width": os.get_terminal_size()[0]}
except OSError:
    _exc_formatter_options = {}


LOGGING_FORMATTERS = {
    "json": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.processors.JSONRenderer(),
        "foreign_pre_chain": shared_processors,
    },
    "plain": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.dev.ConsoleRenderer(exception_formatter=structlog.dev.plain_traceback, colors=False),
        "foreign_pre_chain": shared_processors,
    },
    "pretty": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.dev.ConsoleRenderer(
            exception_formatter=structlog.dev.RichTracebackFormatter(**_exc_formatter_options)
        ),
        "foreign_pre_chain": shared_processors,
    },
    "keyvalue": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.processors.KeyValueRenderer(key_order=["timestamp", "level", "event", "logger"]),
        "foreign_pre_chain": shared_processors,
    },
}

DEFAULT_LOGGING_FORMAT = "pretty"
LOGGING_FORMAT = os.environ.get("LOGGING_FORMAT", DEFAULT_LOGGING_FORMAT)
if LOGGING_FORMAT not in LOGGING_FORMATTERS:
    LOGGING_FORMAT = DEFAULT_LOGGING_FORMAT

logging_config = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": LOGGING_FORMATTERS,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": LOGGING_FORMAT,
        },
    },
    "root": {
        "handlers": ["console"],
        "level": logging.INFO,
    },
    "loggers": {
        "harp": {"level": os.environ.get("LOGGING_HARP", "INFO")},
        "harp_apps": {"level": os.environ.get("LOGGING_HARP", "INFO")},
        "harp.event_dispatcher": {"level": os.environ.get("LOGGING_HARP_EVENTS", "WARNING")},
        "httpcore": {"level": os.environ.get("LOGGING_HTTP", "WARNING")},  # todo wrap in structlog
        "httpx": {"level": os.environ.get("LOGGING_HTTP", "WARNING")},  # todo wrap in structlog
        "hypercorn.access": {"level": os.environ.get("LOGGING_HYPERCORN_ACCESS", "WARNING")},
        "hypercorn.error": {"level": os.environ.get("LOGGING_HYPERCORN_ERROR", "INFO")},
        "sqlalchemy.engine": {"level": os.environ.get("LOGGING_SQLALCHEMY", "WARNING")},
    },
}


logging.config.dictConfig(logging_config)

structlog.configure(
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    processors=processors,
    cache_logger_on_first_use=True,
)


def get_logger(name, *args: Any, **initial_values: Any) -> Any:
    if name == "env_py":
        name = "__migrations__"
    else:
        pkg, mod = name.rsplit(".", 1)
        if mod in ("__init__", "__main__", "__app__"):
            name = pkg

    return structlog.get_logger(name, *args, **initial_values)


__all__ = ["get_logger"]



"""
SqlAlchemy Storage Extension

"""

from harp import get_logger
from harp.config import Application
from harp.config.events import FactoryBindEvent
from harp.typing.storage import Storage

from .settings import SqlAlchemyStorageSettings
from .storage import SqlAlchemyStorage

logger = get_logger(__name__)


class SqlalchemyStorageApplication(Application):
    settings_namespace = "storage"
    settings_type = SqlAlchemyStorageSettings

    @classmethod
    def supports(cls, settings):
        return settings.get("type", None) == "sqlalchemy"

    @classmethod
    def defaults(cls, settings=None):
        settings = settings if settings is not None else {"type": "sqlalchemy"}

        if cls.supports({"type": "sqlalchemy"} | settings):
            settings.setdefault("type", "sqlalchemy")
            settings.setdefault("url", "sqlite+aiosqlite:///:memory:")
            settings.setdefault("migrate", True)

        return settings

    async def on_bind(self, event: FactoryBindEvent):
        event.container.add_singleton(Storage, SqlAlchemyStorage)



"""
Telemetry application.

Sends anonymous usage statistics on startup.

todo:
- send ping every 24 hours
- send ping on exit ?
- activity type
- license key
- send platform to allow to understand target env types (linux, windows, macos, what python, etc.)
- detect if within container ?
- detect if within kubernetes ?
- detect git/docker/pip/brew install ?

"""

from harp import get_logger
from harp.config import Application
from harp.config.events import FactoryBindEvent, FactoryBoundEvent
from harp_apps.telemetry.manager import TelemetryManager

logger = get_logger(__name__)


class TelemetryApplication(Application):
    def __init__(self, settings=None, /):
        super().__init__(settings)

        # keep a reference for application garbage collection (later)
        self.manager = None

    async def on_bind(self, event: FactoryBindEvent):
        event.container.add_singleton(TelemetryManager)

    async def on_bound(self, event: FactoryBoundEvent):
        self.manager = event.provider.get(TelemetryManager)



from typing import Optional

from .base import BaseSetting, settings_dataclass


@settings_dataclass
class DisabledSettings(BaseSetting):
    enabled: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.enabled is True:
            raise ValueError("Cannot set enabled to True in DisabledSettings")

    def __setattr__(self, name, value):
        if name == "enabled" and value is True:
            raise ValueError("Cannot set enabled to True in DisabledSettings")
        super().__setattr__(name, value)

    def __repr__(self):
        return "disabled"


@settings_dataclass
class DisableableBaseSettings(BaseSetting):
    enabled: Optional[bool] = True

    def __new__(cls, **kwargs):
        _enabled = kwargs.pop("enabled", True)

        # todo better parsing of falsy values
        if isinstance(_enabled, str) and _enabled.lower() in {"no", "false", "0"}:
            _enabled = False

        if not _enabled:
            return DisabledSettings()

        return super().__new__(cls)



"""
Factory configuration, does nothing but can be used to instanciate a server.
"""

import importlib.util
import os
import pprint
from copy import deepcopy
from types import MappingProxyType
from typing import Self, Type

import orjson
from config.ini import INIFile
from config.json import JSONFile
from config.toml import TOMLFile
from rodi import Container
from whistle import IAsyncEventDispatcher

from harp import get_logger
from harp.commandline.server import CommonServerOptions
from harp.config.application import Application
from harp.utils.identifiers import is_valid_dotted_identifier

logger = get_logger(__name__)


def get_application_class_name(name):
    return "".join(map(lambda x: x.title().replace("_", ""), name.rsplit(".", 1)[-1:])) + "Application"


def _resolve_application_aliases(spec):
    if "." not in spec:
        _candidate = ".".join(("harp_apps", spec))
        if importlib.util.find_spec(_candidate):
            return _candidate
    return spec


class Config:
    DEFAULT_APPLICATIONS = [
        "harp_apps.http_client",
        "harp_apps.proxy",
        "harp_apps.dashboard",
        "harp_apps.sqlalchemy_storage",
        "harp_apps.telemetry",
        "harp_apps.janitor",
        "harp_apps.contrib.sentry",  # todo: allow to extend application list in config file without overriding all
    ]

    def __init__(self, settings=None, /, applications=None):
        self._raw_settings = {"applications": applications or []} | (settings or {})
        self._validated_settings = None
        self._debug_applications = set()
        self._application_types = {}
        self._applications = []
        self._disabled_applications = set()

    def __eq__(self, other: Self):
        return self.settings == other.settings

    def __repr__(self):
        return (
            f"<{type(self).__name__}{'' if self._validated_settings else '*'} "
            f"settings={orjson.dumps(self.settings.copy()).decode()}>"
        )

    @property
    def settings(self):
        if self._validated_settings is not None:
            return self._validated_settings
        return self._raw_settings

    @property
    def applications(self):
        if not self._validated_settings:
            raise RuntimeError("Configuration not validated.")
        return self._validated_settings.get("applications", [])

    def reset(self):
        self._validated_settings = None

    def add_defaults(self):
        for application in type(self).DEFAULT_APPLICATIONS:
            self.add_application(application)

    def set(self, key, value):
        if not is_valid_dotted_identifier(key):
            raise ValueError(f"Invalid settings key: {key}")

        self.reset()

        bits: list[str] = key.split(".")
        current = self._raw_settings
        for bit in bits[:-1]:
            current = current.setdefault(bit, {})
        current[bits[-1]] = value

    def add_application(self, name, /, *, debug=False):
        if not is_valid_dotted_identifier(name):
            raise ValueError(f"Invalid application name: {name}")
        self.reset()
        name = _resolve_application_aliases(name)
        if name not in self._raw_settings["applications"]:
            self._raw_settings["applications"].append(name)
            if debug:
                self._debug_applications.add(name)

    def remove_application(self, name):
        if not is_valid_dotted_identifier(name):
            raise ValueError(f"Invalid application name: {name}")
        self.reset()
        name = _resolve_application_aliases(name)
        if name in self._raw_settings["applications"]:
            self._raw_settings["applications"].remove(name)
        self._debug_applications.discard(name)

    def read_env(self, options: CommonServerOptions, /):
        """
        Parses sys.argv-like arguments.

        :param args:
        :param files: list of filenames to load in order. Will happen after defaults env and files.
        :return: argparse.Namespace

        """
        for _enabled_application in options.enable or ():
            self.add_application(_enabled_application)
        for _disabled_application in options.disable or ():
            self.remove_application(_disabled_application)

        from config.common import ConfigurationBuilder, MapSource
        from config.env import EnvVars
        from config.yaml import YAMLFile

        builder = ConfigurationBuilder()

        # default config
        builder.add_source(MapSource({}))
        builder.add_source(EnvVars(prefix="DEFAULT__HARP_"))

        # current
        builder.add_source(MapSource(self._raw_settings))

        # load default system config (if present)
        if os.path.exists("/etc/harp.yaml"):
            builder.add_source(YAMLFile("/etc/harp.yaml"))
        elif os.path.exists("/etc/harp.yml"):
            builder.add_source(YAMLFile("/etc/harp.yml"))

        # load user config
        for file in options.files or ():
            _, ext = os.path.splitext(file)
            if ext in (".yaml", ".yml"):
                builder.add_source(YAMLFile(file))
            elif ext in (".json",):
                builder.add_source(JSONFile(file))
            elif ext in (".ini", ".conf"):
                builder.add_source(INIFile(file))
            elif ext in (".toml",):
                builder.add_source(TOMLFile(file))
            else:
                raise ValueError(f"Unknown file extension: {ext}")

        builder.add_source(EnvVars(prefix="HARP_"))

        for k, v in (options.options or {}).items():
            builder.add_value(k, v)

        # reset option
        if options.reset:
            builder.add_value("storage.drop_tables", True)

        self._raw_settings = builder.build().values

        self._raw_settings.setdefault("proxy", {})
        self._raw_settings["proxy"].setdefault("endpoints", [])

        for k, v in (options.endpoints or {}).items():
            _port, _url = v.split(":", 1)
            self._raw_settings["proxy"]["endpoints"].append(
                {
                    "name": k,
                    "port": int(_port),
                    "url": _url,
                }
            )

    def validate(self, *, allow_extraneous_settings=False):
        if self._validated_settings is None:
            to_be_validated = deepcopy(self._raw_settings)
            application_names = to_be_validated.pop("applications", [])
            validated = {"applications": []}

            # round 1: import applications and set defaults before validation
            application_types, to_be_validated = self._validate_round_1_import_applications(
                application_names, to_be_validated
            )
            for application in application_types:
                validated["applications"].append(application.name)

            applications, newly_validated = self._validate_round_2_extract_and_validate_settings(
                application_types, to_be_validated
            )
            newly_validated.pop("applications", None)  # not allowed xxx todo raise
            validated |= newly_validated

            if to_be_validated != {} and not allow_extraneous_settings:
                logger.warning(f"Unknown settings remaining: {to_be_validated}")

            # propagate for the world to use
            self._validated_settings = MappingProxyType(validated)
            self._applications = applications

        logger.debug(f"Configuration validated: {pprint.pformat(self._validated_settings)}")

        return self._validated_settings

    def get_application_type(self, name: str) -> Type[Application]:
        if name not in self._application_types:
            application_spec = importlib.util.find_spec(name)
            if not application_spec:
                raise ValueError(f'Unable to find application "{name}".')

            try:
                application_module = __import__(".".join((application_spec.name, "__app__")), fromlist=["*"])
            except ModuleNotFoundError as exc:
                raise ModuleNotFoundError(
                    f'A python package for application "{name}" was found but it is not a valid harp application. '
                    'Did you forget to add an "__app__.py"?'
                ) from exc

            application_class_name = get_application_class_name(name)
            if not hasattr(application_module, application_class_name):
                raise AttributeError(
                    f'Application module for {name} does not contain a "{application_class_name}" class.'
                )

            self._application_types[application_spec.name] = getattr(application_module, application_class_name)
            self._application_types[application_spec.name].name = application_spec.name

        if name not in self._application_types:
            raise RuntimeError(f'Unable to load application "{name}", application class definition not found.')

        return self._application_types[name]

    def _validate_round_1_import_applications(self, names, to_be_validated):
        # note: to_be_validated is modified in place (is this right ?), but still returned to avoid confusion
        if len(self._debug_applications):
            print("DEBUG: Configuration > Round 1 (import applications) > START")

        types = []

        for _name in names:
            _type = self.get_application_type(_name)

            if _type.settings_namespace:
                _settings = to_be_validated.get(_type.settings_namespace, None)
                if (_defaults := _type.defaults(deepcopy(_settings) if _settings else None)) is not None:
                    to_be_validated[_type.settings_namespace] = _defaults
            else:
                pass  # todo assert application explicitely opted in for no defaults ?

            if _type.name in self._debug_applications:
                print(f"DEBUG: {_type}::set_default_settings(...)")

            types.append(_type)

        if len(self._debug_applications):
            print("DEBUG: Configuration > Round 1 (import applications) > END")
            print(f"DEBUG: to_be_validated={pprint.pformat(to_be_validated)}")

        return types, to_be_validated

    def _validate_round_2_extract_and_validate_settings(self, application_types, to_be_validated):
        applications = []
        validated = {}
        # round 2: validate and extract settings
        if len(self._debug_applications):
            print("DEBUG: Configuration > Round 2 > START")
        for application_type in application_types:
            if application_type.settings_namespace:
                if not application_type.supports(to_be_validated.get(application_type.settings_namespace, {})):
                    raise ValueError(
                        f'Application "{application_type.name}" is not configurable with the current configuration.'
                    )
                application_settings = to_be_validated.pop(application_type.settings_namespace, None)
                if application_type.settings_type:
                    application_settings = application_type.settings_type(**application_settings)
            else:
                application_settings = None

            if application_type.name in self._debug_applications:
                print(f"DEBUG: {application_type}::extract_settings() -> {application_settings}")

            if application_settings is None and application_type.settings_namespace is not None:
                raise ValueError(
                    f'Application "{application_type.name}" is not configurable with the current configuration '
                    f"(expected some configuration, got none)."
                )

            application = application_type(application_settings)
            application_validated_settings = application.validate()
            if application_type.settings_namespace:
                validated |= {application_type.settings_namespace: application_validated_settings}
            applications.append(application)

        if len(self._debug_applications):
            print("DEBUG: Configuration > Round 2 > END")
        return applications, validated

    def register_events(self, dispatcher: IAsyncEventDispatcher):
        self.validate()
        for application in self._applications:
            application.register_events(dispatcher)

    def register_services(self, container: Container):
        self.validate()
        for application in self._applications:
            application.register_services(container)

    def serialize(self):
        self.validate()
        return orjson.dumps(self.settings.copy())

    @classmethod
    def deserialize(cls, settings):
        return cls(orjson.loads(settings))



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from click.testing import CliRunner
from sqlalchemy import URL, make_url

from harp.commandline import migrations


async def run_cli_migrate_command(url: Union[str | URL], /, *, operation="up", revision="head"):
    """
    Run the `migrate` command from the CLI, synchronously, in a thread (via ThreadPoolExecutor). This uses click's
    CliRunner which is a testing helper. You should not use this outside tests, as it will setup a fully separate harp
    instance with its own configuration, etc.

    :param url:
    :param operation: up or down
    :param revision: target revision ("head" for latest, "base" for initial or specific revision number)
    :return:
    """
    url = make_url(url)

    def _migrate():
        cli = CliRunner()
        return cli.invoke(
            migrations.migrate,
            [operation, revision, "--set", "storage.url=" + url.render_as_string(hide_password=False)],
        )

    with ThreadPoolExecutor() as executor:
        result = await asyncio.get_event_loop().run_in_executor(executor, _migrate)

    return result



from contextlib import asynccontextmanager
from datetime import UTC, datetime

import pytest
from sqlalchemy import make_url, text
from sqlalchemy.ext.asyncio import create_async_engine
from whistle import AsyncEventDispatcher

from harp.models import Blob, Message, Transaction
from harp.utils.guids import generate_transaction_id_ksuid
from harp_apps.sqlalchemy_storage.settings import SqlAlchemyStorageSettings
from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage


async def _create_storage(*, settings, dispatcher=None) -> SqlAlchemyStorage:
    storage = SqlAlchemyStorage(
        dispatcher=dispatcher or AsyncEventDispatcher(),
        settings=SqlAlchemyStorageSettings(**settings),
    )
    await storage.initialize()

    return storage


@asynccontextmanager
async def get_scoped_database_url(database_url, test_id):
    admin_url = make_url(database_url)
    dialect = admin_url.get_dialect().name

    if dialect == "mysql":
        admin_url = admin_url.set(username="root")

    admin_engine = create_async_engine(admin_url.render_as_string(hide_password=False))

    if dialect == "sqlite":
        try:
            yield database_url
        finally:
            await admin_engine.dispose()

    elif dialect in ("mysql", "postgresql"):
        db_name = f"test_{test_id}"
        url = make_url(database_url).set(database=db_name)

        async with admin_engine.connect() as conn:
            await conn.execution_options(isolation_level="AUTOCOMMIT")

            await conn.execute(text(f"CREATE DATABASE {url.database}"))
            if dialect == "mysql":
                await conn.execute(text(f"GRANT ALL PRIVILEGES ON {url.database}.* TO '{url.username}'@'%'"))

        try:
            yield url.render_as_string(hide_password=False)
        finally:
            async with admin_engine.connect() as conn:
                await conn.execution_options(isolation_level="AUTOCOMMIT")
                await conn.execute(text(f"DROP DATABASE {url.database}"))
            await admin_engine.dispose()

    else:
        raise RuntimeError(f"Unsupported dialect «{admin_engine.dialect.name}»")


class SqlalchemyStorageTestFixtureMixin:
    storage_settings = {
        "url": "sqlite+aiosqlite:///:memory:",
    }

    def get_sqlalchemy_storage_settings(self, settings):
        return {**self.storage_settings, **settings}

    @pytest.fixture
    async def storage(self, database_url, test_id) -> SqlAlchemyStorage:
        async with get_scoped_database_url(database_url, test_id) as scoped_database_url:
            storage = await _create_storage(
                settings=self.get_sqlalchemy_storage_settings({"url": scoped_database_url}),
            )
            try:
                yield storage
            finally:
                await storage.engine.dispose()

    async def create_transaction(self, storage: SqlAlchemyStorage, **kwargs):
        return await storage.transactions.create(
            Transaction(
                **{
                    **{
                        "id": generate_transaction_id_ksuid(),
                        "type": "http",
                        "endpoint": "/",
                        "started_at": datetime.now(UTC),
                    },
                    **kwargs,
                }
            )
        )

    async def create_blob(self, storage: SqlAlchemyStorage, data, /, **kwargs):
        return await storage.blobs.create(Blob.from_data(data, **kwargs))

    async def create_message(self, storage: SqlAlchemyStorage, **kwargs):
        return await storage.messages.create(Message(**kwargs))



import dataclasses
from datetime import datetime
from typing import List

from .base import Entity
from .messages import Message


@dataclasses.dataclass(kw_only=True)
class Transaction(Entity):
    id: str = None
    """Unique identifier for this transaction."""

    type: str  # enum http websocket lifecycle ...
    """Type of ASGI transaction: http, websocket, lifecycle, ..."""

    endpoint: str = None
    """Endpoint name, if any (this describes which proxy controller handled the request)."""

    started_at: datetime
    """Timestamp of the transaction start."""

    finished_at: datetime = None
    """Timestamp of the transaction end (if it has ended), or None."""

    # Computed fields

    elapsed: float = None
    """Elapsed time in seconds, if the transaction has ended."""

    apdex: int = None
    """Apdex score, if the transaction has ended."""

    # Relations

    messages: List[Message] = None
    tags: dict = dataclasses.field(default_factory=dict)

    # Extra attributes

    extras: dict = dataclasses.field(default_factory=dict)



"""
The Core (:mod:`harp`) package is the root namespace of the Harp framework.

It mostly contains a reference to the :class:`Config` class, because it's the only object you need to start using Harp
using the python API (you don't *need* to use this API, configuration files should be enough for most use cases, but
if you want to, this is the starting point).

For convenience, the :func:`run` function is also available, which is a simple way to start the default server
implementation for your configuration object.

Example usage:

.. code-block:: python

    from harp import Config, run

    config = Config()
    config.add_defaults()

    if __name__ == "__main__":
        run(config)

You can find more information about how configuration works in the :mod:`harp.config` module.

Contents
--------

"""

import os
from subprocess import check_output

from packaging.version import InvalidVersion, Version

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def _parse_version(version: str, /, *, default=None) -> Version:
    try:
        return Version(version)
    except InvalidVersion:
        if "-" in version:
            return _parse_version(version.rsplit("-", 1)[0], default=default)
        return default


# last release
__title__ = "Core"
__version__ = "0.5.0b12"
__revision__ = __version__  # we can't commit the not yet known revision

# override with version.txt if available (after docker build for example)
if os.path.exists(os.path.join(ROOT_DIR, "version.txt")):
    with open(os.path.join(ROOT_DIR, "version.txt")) as f:
        __version__ = f.read().strip()

__parsed_version__ = _parse_version(__version__)

# override with current development version/revision if available (disabled in CI, for docs)
if not os.environ.get("CI", False) and os.path.exists(os.path.join(ROOT_DIR, ".git")):
    __revision__ = check_output(["git", "rev-parse", "HEAD"], cwd=ROOT_DIR).decode("utf-8").strip()
    try:
        __version__ = (
            check_output(["git", "describe", "--tags", "--always", "--dirty"], cwd=ROOT_DIR).decode("utf-8").strip()
        )
        __parsed_version__ = _parse_version(__version__, default=__parsed_version__)
    except Exception:
        __version__ = __revision__[:7]

from ._logging import get_logger  # noqa: E402, isort: skip
from harp.config import Config  # noqa: E402, isort: skip


def run(config: Config):
    """
    Run the default server using provided configuration.

    :param config: Config
    :return:
    """
    import asyncio

    from harp.config.adapters.hypercorn import HypercornAdapter
    from harp.config.factories.kernel_factory import KernelFactory

    factory = KernelFactory(config)
    server = HypercornAdapter(factory)
    return asyncio.run(server.serve())


__all__ = [
    "Config",
    "ROOT_DIR",
    "__revision__",
    "__version__",
    "__parsed_version__",
    "get_logger",
    "run",
]



import asyncio
from inspect import iscoroutinefunction

from harp import get_logger

logger = get_logger(__name__)


class AsyncWorkerQueue:
    def __init__(self):
        self._queue = asyncio.Queue()
        self._task = asyncio.create_task(self())

    async def __call__(self):
        while True:
            try:
                item, ignore_errors = await self._queue.get()
            except RuntimeError:
                # queue is closed
                break
            try:
                await item()
            except Exception as e:
                if not ignore_errors:
                    logger.exception(f"Error while executing queued task: {e}")
            finally:
                self._queue.task_done()

    async def push(self, item, /, *, ignore_errors=False):
        if not iscoroutinefunction(item):
            raise ValueError(f"Unknown item type: {type(item)}, expecting coroutine function.")

        await self._queue.put((item, ignore_errors))

    async def wait_until_empty(self):
        await self._queue.join()



import copy
import dataclasses
from dataclasses import dataclass

from .lazy import Definition, Lazy

settings_dataclass = dataclass


def asdict(obj):
    if hasattr(obj, "_asdict"):
        return obj._asdict()

    if type(obj) in dataclasses._ATOMIC_TYPES:
        return obj

    if dataclasses._is_dataclass_instance(obj):
        # fast path for the common case
        return {f.name: asdict(getattr(obj, f.name)) for f in dataclasses.fields(obj)}

    if isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[asdict(v) for v in obj])

    if isinstance(obj, (list, tuple)):
        return type(obj)(asdict(v) for v in obj)

    if isinstance(obj, dict):
        if hasattr(type(obj), "default_factory"):
            # obj is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = type(obj)(getattr(obj, "default_factory"))
            for k, v in obj.items():
                result[asdict(k)] = asdict(v)
            return result
        return type(obj)((asdict(k), asdict(v)) for k, v in obj.items())

    return copy.deepcopy(obj)


@settings_dataclass
class BaseSetting:
    def to_dict(self):
        return asdict(self)

    def __post_init__(self):
        for _name, _hint in self.__annotations__.items():
            try:
                is_definition = issubclass(_hint, Definition)
            except TypeError:
                is_definition = False

            if hasattr(_hint, "__origin__"):
                try:
                    if issubclass(_hint.__origin__, Definition):
                        is_definition = True
                except TypeError:
                    pass

            if is_definition:
                setattr(self, _name, Lazy(getattr(self, _name), _default=getattr(type(self), _name, None)))



from harp.typing.storage import Storage


class AbstractFacet:
    name = None
    choices = set()
    exhaustive = True

    def __init__(self):
        self.meta = {}

    @property
    def values(self):
        return [{"name": choice, "count": self.meta.get(choice, {}).get("count", None)} for choice in self.choices]

    def get_filter(self, raw_data: list):
        query_endpoints = self.choices.intersection(raw_data)
        return list(query_endpoints) if len(query_endpoints) else None

    def filter(self, raw_data: list):
        return {
            "values": self.values,
            "current": self.get_filter(raw_data),
        }


class FacetWithStorage(AbstractFacet):
    def __init__(self, *, storage: Storage):
        super().__init__()
        self.storage = storage


class NonExhaustiveFacet(AbstractFacet):
    exhaustive = False

    def __init__(self):
        super().__init__()
        self.choices.add("NULL")



from dataclasses import dataclass, field
from itertools import chain
from shlex import quote
from typing import Iterable

from harp.utils.commandline import click, code


@dataclass(kw_only=True)
class CommonServerOptions(dict):
    """
    Common server options, in a dataclass.
    """

    options: dict = field(default_factory=dict)
    endpoints: Iterable = field(default_factory=dict)
    files: tuple = ()
    enable: tuple = ()
    disable: tuple = ()

    # TODO maybe run reset as a pre-start command, so it does not run on each reload?
    reset: bool = False

    def as_list(self):
        return list(
            chain(
                (f"--set {quote(key)}={quote(value)}" for key, value in self.options.items()),
                (f"--endpoint {quote(key)}={quote(value)}" for key, value in self.endpoints.items()),
                ("--enable {app}".format(app=app) for app in self.enable),
                ("--disable {app}".format(app=app) for app in self.disable),
                ("--file " + file for file in self.files),
                (("--set storage.drop_tables=true",) if self.reset else ()),
            )
        )

    def __post_init__(self):
        self.options = dict(map(lambda x: x.split("=", 1), self.options))
        self.endpoints = dict(map(lambda x: x.split("=", 1), self.endpoints))


def add_harp_server_click_options(f):
    """
    Decorate a click command to add common server options, in the right order.
    """
    options = [
        click.option(
            "--set",
            "options",
            multiple=True,
            help=f"Set proxy configuration options (e.g. {code('--set foo=bar')}, can be used multiple times).",
        ),
        click.option(
            "--endpoint",
            "endpoints",
            multiple=True,
            help=f"""Add an endpoint (e.g. {code('--endpoint httpbin=4000:http://httpbin.org/')}, can be used multiple
            times).""",
        ),
        click.option(
            "--file",
            "-f",
            "files",
            default=(),
            multiple=True,
            help="""Load configuration from file (configuration format will be detected from file extension, can be
            used multiple times).""",
        ),
        click.option("--enable", default=(), multiple=True, help="Enable some applications."),
        click.option("--disable", default=(), multiple=True, help="Disable some applications."),
    ]

    # apply options in reversed order so that click will apply them in the right order (it's intended to be used as a
    # decorator, hence the reversal).
    for option in reversed(options):
        f = option(f)

    return f



from datetime import datetime
from typing import Iterable, List, Optional, Protocol, TypedDict


class TransactionsGroupedByTimeBucket(TypedDict):
    datetime: datetime
    count: int
    errors: int
    meanDuration: float


class Storage(Protocol):
    async def get_transaction_list(
        self,
        *,
        username: str,
        with_messages=False,
        filters=None,
        page: int = 1,
        cursor: str = "",
        text_search: str = "",
    ):
        """Find transactions, using optional filters, for example to be displayed in the dashboard."""
        ...

    async def get_transaction(
        self,
        id: str,
        /,
        *,
        username: str,
    ):
        """Find a transaction, by id."""
        ...

    async def get_blob(self, blob_id):
        """Retrieve a blob, by id. Blobs are content adressable, although the hash is built on the source data and
        blob filters may be applied to the source data before storage, meaning that the hash may not be re-computable
        once storage is done."""
        ...

    async def get_facet_meta(self, name):
        """Retrieve a facet's metadata, by name."""
        ...

    async def transactions_grouped_by_time_bucket(
        self,
        *,
        endpoint=None,
        time_bucket: Optional[str],
        start_datetime: Optional[datetime],
    ) -> List[TransactionsGroupedByTimeBucket]: ...

    async def set_user_flag(self, *, transaction_id: str, username: str, flag: int, value=True):
        """Sets or unsets a user flag on a transaction."""
        ...

    async def create_users_once_ready(self, users: Iterable[str]):
        """Create users."""
        ...

    async def get_usage(self):
        """Get storage usage."""
        ...



from tempfile import NamedTemporaryFile

from harp import Config


class TestSettings:
    async def factory(self, settings=None):
        config = Config(settings)
        config.add_application("sqlalchemy_storage")
        config.add_application("dashboard")
        config.validate()
        return config.settings

    async def test_dashboard_auth_basic(self):
        users = {"foo": "bar"}
        settings = await self.factory({"dashboard": {"auth": {"type": "basic", "algorithm": "plain", "users": users}}})
        assert settings["dashboard"]["auth"] == {"algorithm": "plain", "type": "basic", "users": {"foo": "bar"}}

    async def test_dashboard_auth_basic_from_file(self):
        with NamedTemporaryFile("w") as f:
            f.write("foo:bar")
            f.close()
            settings = await self.factory({"dashboard": {"auth": {"type": "basic", "users": {"fromFile": f.name}}}})

        assert settings["dashboard"]["auth"] == {
            "algorithm": "pbkdf2_sha256",
            "type": "basic",
            "users": {"from_file": f.name},
        }



from datetime import UTC, datetime
from functools import cached_property
from typing import Optional
from urllib.parse import urlencode, urljoin, urlparse

import httpx
from hishel._headers import parse_cache_control
from httpx import AsyncClient, codes
from whistle import IAsyncEventDispatcher

from harp import __parsed_version__, get_logger
from harp.asgi.events import MessageEvent, TransactionEvent
from harp.http import BaseHttpMessage, HttpError, HttpRequest, HttpResponse
from harp.http.requests import WrappedHttpRequest
from harp.models import Transaction
from harp.utils.guids import generate_transaction_id_ksuid

from .events import EVENT_TRANSACTION_ENDED, EVENT_TRANSACTION_MESSAGE, EVENT_TRANSACTION_STARTED

logger = get_logger(__name__)


class HttpProxyController:
    name: str = None
    """Controller name, also refered as endpoint name (for example in
    :class:`Transaction <harp.models.Transaction>`)."""

    user_agent: str = None
    """User agent to use when proxying requests (will default to harp/<version>)."""

    _dispatcher: Optional[IAsyncEventDispatcher] = None
    """Event dispatcher for this controller."""

    url: str
    """Base URL to proxy requests to."""

    @cached_property
    def dispatcher(self):
        """Read-only reference to the event dispatcher."""
        return self._dispatcher

    def __init__(self, url, *, http_client: AsyncClient, dispatcher=None, name=None):
        self.http_client = http_client
        self.url = url or self.url
        self.name = name or self.name
        self._dispatcher = dispatcher or self._dispatcher

        self.parsed_url = urlparse(self.url)

        # we only expose minimal information about the exact version
        if not self.user_agent:
            try:
                self.user_agent = f"harp/{__parsed_version__.major}.{__parsed_version__.minor}"
            except AttributeError:
                self.user_agent = "harp"

    async def adispatch(self, event_id, event=None):
        """
        Shortcut method to dispatch an event using the controller's dispatcher, if there is one.

        :return: :class:`IEvent <whistle.IEvent>` or None
        """
        if self._dispatcher:
            return await self._dispatcher.adispatch(event_id, event)

    async def __call__(self, request: HttpRequest):
        """Handle an incoming request and proxy it to the configured URL.

        :param request: ASGI request
        :param response: ASGI response
        :return:

        """

        # create an envelope to override things, without touching the original request
        request = WrappedHttpRequest(request)
        request.headers["host"] = self.parsed_url.netloc

        # BEGIN TRANSACTION AND PREPARE DATA FOR PROXY REQUEST
        tags = {}

        for header in request.headers:
            header = header.lower()
            if header.startswith("x-harp-"):
                tags[header[7:]] = request.headers.pop(header)

        # override user agent (later, may be customizable behavior)
        if self.user_agent:
            request.headers["user-agent"] = self.user_agent

        transaction = await self._create_transaction_from_request(request, tags=tags)
        await request.join()
        url = urljoin(self.url, request.path) + (f"?{urlencode(request.query)}" if request.query else "")

        # PROXY REQUEST
        p_request: httpx.Request = self.http_client.build_request(
            request.method, url, headers=list(request.headers.items()), content=request.body
        )
        logger.debug(f"▶▶ {request.method} {url}", transaction_id=transaction.id)

        # PROXY RESPONSE
        try:
            p_response: httpx.Response = await self.http_client.send(p_request)
        except httpx.ConnectError as exc:
            transaction.extras["status_class"] = "ERR"
            await self.end_transaction(transaction, HttpError("Unavailable", exception=exc))
            # todo add web debug information if we are not on a production env
            return HttpResponse(
                "Service Unavailable (remote server unavailable)", status=503, content_type="text/plain"
            )
        except httpx.TimeoutException as exc:
            transaction.extras["status_class"] = "ERR"
            await self.end_transaction(transaction, HttpError("Timeout", exception=exc))
            # todo add web debug information if we are not on a production env
            return HttpResponse("Gateway Timeout (remote server timeout)", status=504, content_type="text/plain")

        logger.debug(
            f"◀◀ {p_response.status_code} {p_response.reason_phrase} ({p_response.elapsed.total_seconds()}s)",
            transaction_id=transaction.id,
        )

        response_status = p_response.status_code
        response_headers = {
            k: v
            for k, v in p_response.headers.multi_items()
            if k.lower() not in ("server", "date", "content-encoding", "content-length")
        }
        # XXX for now, we use transaction "extras" to store searchable data for later
        transaction.extras["status_class"] = f"{response_status // 100}xx"

        if p_response.extensions.get("from_cache"):
            transaction.extras["cached"] = p_response.extensions.get("cache_metadata", {}).get("cache_key", True)

        response = HttpResponse(p_response.content, status=response_status, headers=response_headers)

        await self.end_transaction(transaction, response)

        return response

    async def end_transaction(self, transaction, response: BaseHttpMessage):
        spent = int((datetime.now(UTC).timestamp() - transaction.started_at.timestamp()) * 100000) / 100
        transaction.finished_at = datetime.now(UTC)
        transaction.elapsed = spent

        if isinstance(response, HttpError):
            logger.error(f"◀ {type(response).__name__} {response.message} ({spent}ms)", transaction_id=transaction.id)
        elif isinstance(response, HttpResponse):
            reason = codes.get_reason_phrase(response.status)
            logger.debug(f"◀ {response.status} {reason} ({spent}ms)", transaction_id=transaction.id)
        else:
            raise ValueError(f"Invalid final message type: {type(response)}")

        # dispatch message event for response
        # TODO delay after response is sent ?
        await self.adispatch(EVENT_TRANSACTION_MESSAGE, MessageEvent(transaction, response))
        # dispatch transaction ended event
        # TODO delay after response is sent ?
        await self.adispatch(EVENT_TRANSACTION_ENDED, TransactionEvent(transaction))

    async def _create_transaction_from_request(self, request: HttpRequest, *, tags=None):
        transaction = Transaction(
            id=generate_transaction_id_ksuid(),
            type="http",
            started_at=datetime.now(UTC),
            endpoint=self.name,
            tags=tags,
        )

        request_cache_control = request.headers.get("cache-control")
        if request_cache_control:
            request_cache_control = parse_cache_control([request_cache_control])
            if request_cache_control.no_cache:
                transaction.extras["no_cache"] = True

        # XXX for now, we use transaction "extras" to store searchable data for later
        transaction.extras["method"] = request.method

        logger.debug(f"▶ {request.method} {request.path}", transaction_id=transaction.id)

        # dispatch transaction started event
        # we don't really want to await this, should run in background ? or use an async queue ?
        await self.adispatch(EVENT_TRANSACTION_STARTED, TransactionEvent(transaction))

        # dispatch message event for request
        await self.adispatch(EVENT_TRANSACTION_MESSAGE, MessageEvent(transaction, request))

        return transaction

    def __repr__(self):
        return f"{type(self).__name__}({self.url!r}, name={self.name!r})"



import re
from copy import deepcopy
from typing import cast

from sqlalchemy import func, select
from sqlalchemy.orm import aliased, joinedload

from harp import __revision__, __version__, get_logger
from harp.controllers import GetHandler, RouterPrefix, RoutingController
from harp.http import HttpRequest
from harp.typing import GlobalSettings, Storage
from harp.views.json import json
from harp_apps.sqlalchemy_storage.models import MetricValue
from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage

from ..utils.dependencies import get_python_dependencies, parse_dependencies

logger = get_logger(__name__)


@RouterPrefix("/api/system")
class SystemController(RoutingController):
    def __init__(self, *, storage: Storage, settings: GlobalSettings, handle_errors=True, router=None):
        # a bit of scrambling for passwords etc.
        if "storage" in settings:
            if "url" in settings["storage"]:
                settings["storage"]["url"] = re.sub(r"//[^@]*@", "//***@", settings["storage"]["url"])

        self.settings = deepcopy(dict(settings))
        self.storage: SqlAlchemyStorage = cast(SqlAlchemyStorage, storage)

        self._dependencies = None

        super().__init__(handle_errors=handle_errors, router=router)

    @GetHandler("/")
    async def get(self, request: HttpRequest):
        context = getattr(request, "context", {})

        return json(
            {
                "version": __version__,
                "revision": __revision__,
                "user": context.get("user"),
            }
        )

    @GetHandler("/settings")
    async def get_settings(self):
        return json(self.settings)

    @GetHandler("/dependencies")
    async def get_dependencies(self):
        return json({"python": await self.__get_cached_python_dependencies()})

    @GetHandler("/storage")
    async def get_storage(self):
        subquery = select(
            func.rank().over(order_by=MetricValue.created_at.desc(), partition_by=MetricValue.metric_id).label("rank"),
            MetricValue,
        ).subquery()
        v = aliased(MetricValue, subquery)
        query = select(v).where(subquery.c.rank == 1).options(joinedload(v.metric))

        async with self.storage.session_factory() as session:
            result = (await session.execute(query)).scalars().all()

        return json(
            {
                "settings": self.settings.get("storage", {}),
                "counts": {value.metric.name.split(".", 1)[-1]: value.value for value in result},
            }
        )

    async def __get_cached_python_dependencies(self):
        if self._dependencies is None:
            self._dependencies = parse_dependencies(await get_python_dependencies())
        return self._dependencies



from harp.utils.urls import extract_hostname_from_url


def test_extract_hostname_from_url():
    assert extract_hostname_from_url("https://www.google.com/") == "www.google.com"
    assert extract_hostname_from_url("https://example.com:1234/") == "example.com"



from dataclasses import is_dataclass
from typing import Optional

from rodi import Container
from whistle import IAsyncEventDispatcher

from .events import EVENT_FACTORY_BIND, EVENT_FACTORY_BOUND, EVENT_FACTORY_BUILD
from .settings import asdict


class Application:
    """
    Base class for writing harp applications, which are basically python packages with auto-registration superpowers.

    TODO:
    - use settings instead of settings_type, with type annotation ?
    - namespace instead of settings namespace ? autodetect from settings ?

    """

    name = None

    settings_namespace = None
    settings_type = None

    on_bind = None
    """
    Placeholder for factory bind event, happening before the container is built. If set, it will be attached to the
    factory dispatcher automatically.

    """

    on_bound = None
    """
    Placeholder for factory bound event, happening after the container is built. If set, it will be attached to the
    factory dispatcher automatically.
    """

    on_build = None
    """
    Placeholder for factory build event, happening after the kernel is built. If set, it will be attached to the
    factory dispatcher automatically.
    """

    def __init__(self, settings=None, /):
        settings = settings or {}
        if isinstance(settings, dict) and self.settings_type is not None:
            settings = self.settings_type(**settings)
        self.settings = settings

    @staticmethod
    def defaults(settings: Optional[dict] = None) -> dict:
        settings = settings if settings is not None else {}
        return settings

    @classmethod
    def supports(cls, settings: dict) -> bool:
        return True

    def validate(self):
        if is_dataclass(self.settings):
            return asdict(self.settings)
        return self.settings

    def register_events(self, dispatcher: IAsyncEventDispatcher):
        if self.on_bind is not None:
            dispatcher.add_listener(EVENT_FACTORY_BIND, self.on_bind)
        if self.on_bound is not None:
            dispatcher.add_listener(EVENT_FACTORY_BOUND, self.on_bound)
        if self.on_build is not None:
            dispatcher.add_listener(EVENT_FACTORY_BUILD, self.on_build)

    def register_services(self, container: Container):
        if self.settings and self.settings_type:
            container.add_instance(self.settings, self.settings_type)

    def __repr__(self):
        return f'<{type(self).__name__} name="{self.name}">'



from harp.commandline.server import CommonServerOptions


def test_default():
    options = CommonServerOptions(options=(), files=(), enable=(), disable=(), reset=False)
    assert options.as_list() == []


def test_applications():
    options = CommonServerOptions(options=(), files=(), enable=("foo", "bar"), disable=("baz", "blurp"), reset=False)
    assert options.as_list() == ["--enable foo", "--enable bar", "--disable baz", "--disable blurp"]


def test_reset():
    options = CommonServerOptions(options=(), files=(), enable=(), disable=(), reset=True)
    assert options.as_list() == [
        "--set storage.drop_tables=true",
    ]



from typing import TYPE_CHECKING

from harp import get_logger

from .default import not_found_controller
from .typing import ControllerResolver

if TYPE_CHECKING:
    from harp.http import HttpRequest

logger = get_logger(__name__)


class DefaultControllerResolver(ControllerResolver):
    def __init__(self, *, default_controller=None):
        self.default_controller = default_controller or not_found_controller

    async def resolve(self, request: "HttpRequest"):
        return self.default_controller


class ProxyControllerResolver(DefaultControllerResolver):
    def __init__(self, *, default_controller=None):
        super().__init__(default_controller=default_controller)
        self._ports = {}

    @property
    def ports(self):
        return tuple(self._ports.keys())

    def add(self, port: int | str, controller):
        port = int(port)
        self._ports[port] = controller
        logger.info(f"🏭 Map: *:{port} -> {controller}")

    async def resolve(self, request: "HttpRequest"):
        return self._ports.get(request.server_port, self.default_controller)



import hishel
import httpx

from harp.config import DisabledSettings
from harp_apps.http_client.client import AsyncHttpClient
from harp_apps.http_client.settings import CacheSettings, HttpClientSettings


class TestHttpClientSettings:
    def test_without_cache(self):
        settings = HttpClientSettings(cache={"enabled": False})

        assert settings.cache.enabled is False
        assert isinstance(settings.cache, DisabledSettings)

        client = AsyncHttpClient(settings)
        assert isinstance(client._transport, httpx.AsyncHTTPTransport)

    def test_without_cache_with_custom_client(self):
        settings = HttpClientSettings(
            transport={"@type": "httpx._client:BaseClient"},
            cache={"enabled": False},
        )

        assert settings.cache.enabled is False
        assert isinstance(settings.cache, DisabledSettings)

        client = AsyncHttpClient(settings)
        assert isinstance(client._transport, httpx._client.BaseClient)

    def test_with_default_cache(self):
        settings = HttpClientSettings()

        assert settings.cache.enabled is True
        assert isinstance(settings.cache, CacheSettings)

        client = AsyncHttpClient(settings)
        assert isinstance(client._transport, hishel.AsyncCacheTransport)
        assert client._transport._controller._allow_heuristics is True
        assert client._transport._controller._allow_stale is True
        assert client._transport._controller._cacheable_methods == ["GET", "HEAD"]
        assert client._transport._controller._cacheable_status_codes == list(
            hishel.HEURISTICALLY_CACHEABLE_STATUS_CODES
        )

    def test_with_custom_cache(self):
        settings = HttpClientSettings(
            cache={
                "enabled": True,
                "controller": {
                    "allow_heuristics": False,
                    "cacheable_methods": ["GET"],
                    "cacheable_status_codes": [200],
                },
            }
        )

        assert settings.cache.enabled is True
        assert isinstance(settings.cache, CacheSettings)

        client = AsyncHttpClient(settings)
        assert isinstance(client._transport, hishel.AsyncCacheTransport)

        assert isinstance(client._transport._controller, hishel.Controller)
        assert client._transport._controller._allow_heuristics is False
        assert client._transport._controller._allow_stale is True
        assert client._transport._controller._cacheable_methods == ["GET"]
        assert client._transport._controller._cacheable_status_codes == [200]



from dataclasses import dataclass
from functools import cached_property
from importlib import import_module
from importlib.util import find_spec
from typing import Any, Generic, Mapping, Optional, Tuple, TypeVar

T = TypeVar("T")


@dataclass(kw_only=True, frozen=True)
class Definition(Generic[T]):
    path: str
    name: str
    args: Optional[Tuple] = None
    kwargs: Optional[Mapping] = None

    @cached_property
    def spec(self):
        return find_spec(self.path)

    @cached_property
    def module(self):
        return import_module(self.path)

    @cached_property
    def factory(self):
        return getattr(self.module, self.name)

    def validate(self):
        # For now, only validate module exists, but we may want to import it to check the existence of the attribute
        # inside. We opted out for now as the import can (but should not) have side effects.
        if not self.spec:
            raise ImportError(f"Module {self.path} not found.")
        return True

    def build(self, *args, **kwargs) -> T:
        return self.factory(*self.args, *args, **self.kwargs, **kwargs)

    def _asdict(self):
        return {
            "@type": ":".join((self.path, self.name)),
            **({"@args": self.args} if len(self.args) else {}),
            **self.kwargs,
        }

    to_dict = _asdict


@dataclass(frozen=True)
class ConstantDefinition(Definition):
    value: Any = None
    path: str = None
    name: str = None

    @cached_property
    def spec(self):
        return None

    @cached_property
    def module(self):
        return None

    @cached_property
    def factory(self):
        return lambda: self.value

    def build(self):
        return self.value

    def _asdict(self):
        return self.value


def Lazy(path_or_factory, *args, _default=None, **kwargs) -> Definition[type]:
    if path_or_factory is None:
        return ConstantDefinition(value=None)

    if isinstance(path_or_factory, Definition):
        return path_or_factory

    if callable(path_or_factory):
        # noinspection PyTypeChecker
        return type(
            path_or_factory.__name__ + "Definition",
            (Definition,),
            {"factory": path_or_factory},
        )(
            path=path_or_factory.__module__,
            name=path_or_factory.__name__,
            args=args,
            kwargs=kwargs,
        )

    if isinstance(path_or_factory, dict):
        path = path_or_factory.pop("@type", _default.factory if _default else None)
        args = path_or_factory.pop("@args", ())
        return Lazy(path, *args, **((_default.kwargs if _default else {}) | path_or_factory))

    if isinstance(path_or_factory, str):
        path, name = path_or_factory.rsplit(":", 1)
        return Definition(path=path, name=name, args=args, kwargs=kwargs)

    raise ValueError(f"Invalid lazy factory definition {repr(path_or_factory)}.")



import asyncio
import os

from asgi_middleware_static_file import ASGIMiddlewareStaticFile
from asgiref.typing import ASGISendCallable
from http_router import NotFoundError
from httpx import AsyncClient

from harp import ROOT_DIR, get_logger
from harp.controllers import RoutingController
from harp.errors import ConfigurationError
from harp.http import AlreadyHandledHttpResponse, HttpRequest, HttpResponse
from harp.typing.global_settings import GlobalSettings
from harp.typing.storage import Storage
from harp_apps.proxy.controllers import HttpProxyController

from ..settings import DashboardAuthBasicSetting, DashboardSettings
from .blobs import BlobsController
from .overview import OverviewController
from .system import SystemController
from .transactions import TransactionsController

logger = get_logger(__name__)

# Static directories to look for pre-built assets, in order of priority.
STATIC_BUILD_PATHS = [
    os.path.realpath(os.path.join(ROOT_DIR, "harp_apps/dashboard/frontend/dist")),
    os.path.realpath(os.path.join(ROOT_DIR, "harp_apps/dashboard/web")),
    os.path.realpath(os.path.join(ROOT_DIR, "frontend/dist")),
    "/opt/harp/public",
]


class DashboardController:
    name = "ui"

    storage: Storage
    settings: DashboardSettings
    global_settings: GlobalSettings
    http_client: AsyncClient

    _ui_static_middleware = None
    _ui_devserver_proxy_controller = None

    def __init__(
        self,
        storage: Storage,
        all_settings: GlobalSettings,
        local_settings: DashboardSettings,
        http_client: AsyncClient,
    ):
        # context for usage in handlers
        self.http_client = http_client
        self.storage = storage
        self.global_settings = all_settings
        self.settings = local_settings

        # create users if they don't exist
        if isinstance(self.settings.auth, DashboardAuthBasicSetting):
            asyncio.create_task(self.storage.create_users_once_ready(self.settings.auth.users))

        # controllers for delegating requests
        if self.settings.devserver_port:
            self._ui_devserver_proxy_controller = self._create_ui_devserver_proxy_controller(
                port=self.settings.devserver_port
            )

        # register the subcontrollers, aka the api handlers
        self._internal_api_controller = self._create_internal_api_controller()

        # if no devserver is configured, we may need to serve static files
        if not self._ui_devserver_proxy_controller:
            for _path in STATIC_BUILD_PATHS:
                if os.path.exists(_path):
                    self._ui_static_middleware = ASGIMiddlewareStaticFile(None, "", [_path])
                    break

        # if no devserver is configured and no static files are found, we can't serve the dashboard
        if not self._ui_static_middleware and not self._ui_devserver_proxy_controller:
            raise ConfigurationError(
                "Dashboard controller could not initiate because it got neither compiled assets nor a devserver "
                "configuration."
            )

    def __repr__(self):
        features = {
            "api": bool(self._internal_api_controller),
            "devserver": bool(self._ui_devserver_proxy_controller),
            "static": bool(self._ui_static_middleware),
        }
        return f"{type(self).__name__}({'+'.join(f for f in features if features[f])})"

    def _create_ui_devserver_proxy_controller(self, *, port):
        return HttpProxyController(f"http://localhost:{port}/", http_client=self.http_client)

    def _create_internal_api_controller(self):
        root = RoutingController(handle_errors=False)

        self.children = [
            BlobsController(storage=self.storage, router=root.router),
            SystemController(storage=self.storage, settings=self.global_settings, router=root.router),
            TransactionsController(storage=self.storage, router=root.router),
            OverviewController(storage=self.storage, router=root.router),
        ]

        return root

    async def __call__(self, request: HttpRequest, asgi_send: ASGISendCallable, *, transaction_id=None):
        request.context.setdefault("user", None)

        if self.settings.auth:
            current_auth = request.basic_auth

            if current_auth:
                request.context["user"] = self.settings.auth.check(current_auth[0], current_auth[1])

            if not request.context["user"]:
                return HttpResponse(
                    b"Unauthorized",
                    status=401,
                    headers={"WWW-Authenticate": 'Basic realm="Harp Dashboard"'},
                    content_type="text/plain",
                )

        # Is this a prebuilt static asset?
        if self._ui_static_middleware and not request.path.startswith("/api/"):
            # XXX todo fix
            await self._ui_static_middleware(
                {
                    "type": "http",
                    "path": request.path if "." in request.path else "/index.html",
                    "method": request.method,
                },
                request._impl.asgi_receive,
                asgi_send,
            )
            return AlreadyHandledHttpResponse()

        try:
            return await self._internal_api_controller(request)
        except NotFoundError:
            if self._ui_devserver_proxy_controller:
                return await self._ui_devserver_proxy_controller(request)

        return HttpResponse("Not found.", status=404, content_type="text/plain")



from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage
from harp_apps.sqlalchemy_storage.utils.testing.mixins import SqlalchemyStorageTestFixtureMixin


class TestStorageTransactions(SqlalchemyStorageTestFixtureMixin):
    async def test_get_transaction_list_with_tags(self, storage: SqlAlchemyStorage):
        t1 = await self.create_transaction(storage, endpoint="foo")

        t2 = await self.create_transaction(storage, endpoint="bar")

        t3 = await self.create_transaction(storage, endpoint="baz")

        # messages
        await self.create_message(storage, transaction_id=t1.id, kind="misc", summary="bal", headers="foo", body="foo")
        await self.create_message(storage, transaction_id=t2.id, kind="misc", summary="foo", headers="bar", body="baz")
        await self.create_message(storage, transaction_id=t3.id, kind="misc", summary="baz", headers="baz", body="baz")

        # assert stuff
        transactions_bar = await storage.get_transaction_list(
            username="anonymous", with_messages=True, text_search="bar"
        )
        assert len(transactions_bar) == 1

        assert transactions_bar[0].id == t2.id

        transactions_fo = await storage.get_transaction_list(username="anonymous", with_messages=True, text_search="fo")
        assert len(transactions_fo) == 2

        transactions_ba = await storage.get_transaction_list(username="anonymous", with_messages=True, text_search="ba")
        assert len(transactions_ba) == 3



from harp.commandline.options.server import CommonServerOptions, add_harp_server_click_options
from harp.utils.commandline import click


@click.command(
    short_help="Starts HARP server.",
    help="""Starts HARP server, using the provided configuration. This is the main process and will be the only process
    you need on a live server, it will serve both the proxy ports and the compiled frontend assets (dashboard).""",
)
@add_harp_server_click_options
def server(**kwargs):
    options = CommonServerOptions(**kwargs)

    from harp import Config, run

    config = Config()
    config.add_defaults()
    config.read_env(options)

    return run(config)



import asyncio
from typing import cast

from harp import get_logger
from harp.typing import Storage
from harp_apps.sqlalchemy_storage.storage import SqlAlchemyStorage

from ..sqlalchemy_storage.models.base import with_session
from .settings import OLD_AFTER, PERIOD

logger = get_logger(__name__)


class JanitorWorker:
    def __init__(self, storage: Storage):
        self.storage: SqlAlchemyStorage = cast(SqlAlchemyStorage, storage)
        self.running = False
        self.session_factory = self.storage.session_factory

    def stop(self):
        """
        Mark the loop for termination.
        """
        self.running = False

    async def run(self):
        """
        Once dependencies are ready, start the main loop (basically, run the `loop()` every PERIOD seconds), until
        `stop()` is called.
        """
        # do not start before storage is ready
        await self.storage.ready()

        self.running = True
        while self.running:
            try:
                await self.loop()
            except Exception as exc:
                logger.exception(exc)
            await asyncio.sleep(PERIOD)

    async def loop(self):
        """
        One iteration of the janitor loop.
        """

        # Delete old transactions
        result = await self.delete_old_transactions()
        if result.rowcount:
            logger.debug("🧹 Deleted %d old transactions", result.rowcount)

        # Delete orphan blobs
        result = await self.delete_orphan_blobs()
        if result.rowcount:
            logger.debug("🧹 Deleted %d orphan blobs", result.rowcount)

        # Compute and store stored objecg counts as metrics
        logger.debug("🧹 Compute and store metrics...")
        await self.compute_and_store_metrics()

    @with_session
    async def delete_old_transactions(self, /, *, session):
        """
        Remove transactions older than OLD_AFTER days. On correct database implementations (postgresql for example), it
        will cascade to related objects. On sqlite, there will be garbage left, but it's not a big deal.
        """
        result = await session.execute(self.storage.transactions.delete_old(OLD_AFTER))
        await session.commit()
        return result

    @with_session
    async def delete_orphan_blobs(self, /, *, session):
        """
        Find and remove blobs that are not referenced anymore by any transaction.
        """
        result = await session.execute(self.storage.blobs.delete_orphans())
        await session.commit()
        return result

    @with_session
    async def compute_and_store_metrics(self, /, *, session):
        """
        Compute counts of objects in storage, and store them as metrics.
        """
        await self.storage.metrics.insert_values(await self.compute_metrics(session))

    async def compute_metrics(self, session):
        return {
            "storage.transactions": await self.do_count(session, "transactions"),
            "storage.messages": await self.do_count(session, "messages"),
            "storage.blobs": await self.do_count(session, "blobs"),
            "storage.blobs.orphans": await self.do_count(session, "blobs", method="count_orphans"),
        }

    async def do_count(self, session, name: str, /, *, method="count"):
        """
        Helper to count objects in storage, from different repositories and using different methods for building the
        actual query.

        :param session: sqlalchemy async session
        :param name: repository name (should be available from storage)
        :param method: method name to call on the repository to get the actual sqlalchemy query, default as "count"
        :return: integer
        """
        return (
            await session.execute(
                getattr(
                    getattr(self.storage, name),
                    method,
                )()
            )
        ).scalar()



from functools import cached_property
from typing import Literal, Optional

from harp.config import BaseSetting, FromFileSetting, settings_dataclass
from harp.errors import ConfigurationRemovedSettingError, ConfigurationValueError


def _plain(actual, expected):
    return actual == expected


@settings_dataclass
class DashboardAuthBasicSetting(BaseSetting):
    type: str = "basic"
    algorithm: str = "pbkdf2_sha256"
    users: Optional[FromFileSetting | dict[str, str]] = None

    def __post_init__(self):
        FromFileSetting.may_override(self, "users")

    @cached_property
    def algorithm_impl(self):
        if self.algorithm == "plain":
            return _plain
        # todo check we do not import anything else than a valid algorithm implementation
        impl = __import__("passlib.hash", fromlist=[self.algorithm])
        return getattr(impl, self.algorithm).verify

    def check(self, username, password):
        if not self.users:
            return False

        if isinstance(self.users, dict):
            user = self.users.get(username)
            if not user:
                return False
            if "password" not in user:
                return False
            if not self.algorithm_impl(password, user["password"]):
                return False
            return username
        else:
            raise NotImplementedError("Only dict is supported for now")


@settings_dataclass
class DashboardAuthSetting(BaseSetting):
    type: Optional[Literal["basic"] | None] = None

    def __new__(cls, **kwargs):
        _type = kwargs.pop("type", None)
        available_types = ("basic",)

        if _type is None or not _type:
            if len(kwargs) > 0:
                raise ConfigurationValueError(
                    "Invalid configuration for dashboard auth. There should not be additional arguments if no auth "
                    f"type is provided. Available types: {', '.join(available_types)}."
                )
            return None

        if _type == "basic":
            return DashboardAuthBasicSetting(**kwargs)

        raise ConfigurationValueError(
            f"Invalid dashboard auth type: {_type}. Available types: {', '.join(available_types)}."
        )

    def check(self, username, password):
        raise NotImplementedError()


@settings_dataclass
class DashboardSettings(BaseSetting):
    """Root settings for the dashboard."""

    port: int | str = 4080
    auth: Optional[DashboardAuthSetting] = None
    devserver_port: Optional[int | str] = None

    def __post_init__(self):
        if isinstance(self.auth, str):
            raise ConfigurationRemovedSettingError(
                "Invalid configuration for dashboard auth: string configuration is not supported anymore."
            )

        if isinstance(self.auth, dict):
            object.__setattr__(self, "auth", DashboardAuthSetting(**self.auth))



import http

import pytest

from harp.http.utils import HTTP_METHODS
from harp.typing import Maybe, NotSet


def parametrize_with_http_status_codes(include=None):
    statuses = list(map(lambda x: x.value, http.HTTPStatus))
    if include is not None:
        statuses = list(filter(lambda x: x // 100 in include, statuses))
    return pytest.mark.parametrize("status_code", statuses)


def parametrize_with_http_methods(
    *,
    include_safe=NotSet,
    include_unsafe=NotSet,
    include_idempotent=NotSet,
    include_non_idempotent=NotSet,
    include_standard=NotSet,
    include_non_standard=NotSet,
    include_having_request_body=NotSet,
    include_maybe_having_request_body=NotSet,
    include_not_having_request_body=NotSet,
    include_having_response_body=NotSet,
    include_maybe_having_response_body=NotSet,
    include_not_having_response_body=NotSet,
    exclude=(),
):
    if {
        include_safe,
        include_unsafe,
        include_idempotent,
        include_non_idempotent,
        include_standard,
        include_non_standard,
        include_having_request_body,
        include_maybe_having_request_body,
        include_not_having_request_body,
        include_having_response_body,
        include_maybe_having_response_body,
        include_not_having_response_body,
    } == {NotSet}:
        include_standard = True

    methods = set()
    if include_safe is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.safe is True}
    if include_unsafe is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.safe is False}
    if include_idempotent is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.idempotent is True}
    if include_non_idempotent is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.idempotent is False}
    if include_standard is True:
        methods |= set(HTTP_METHODS.keys())
    if include_non_standard is True:
        methods |= {"BREW", "REMIX"}
    if include_having_request_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.request_body is True}
    if include_not_having_request_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.request_body is False}
    if include_maybe_having_request_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.request_body is Maybe}
    if include_having_response_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.response_body is True}
    if include_not_having_response_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.response_body is False}
    if include_maybe_having_response_body is True:
        methods |= {name for name, method in HTTP_METHODS.items() if method.response_body is Maybe}

    if include_safe is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.safe is True}
    if include_unsafe is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.safe is False}
    if include_idempotent is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.idempotent is True}
    if include_non_idempotent is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.idempotent is False}
    if include_standard is False:
        methods -= set(HTTP_METHODS.keys())
    if include_non_standard is False:
        methods -= {"BREW", "REMIX"}
    if include_having_request_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.request_body is True}
    if include_not_having_request_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.request_body is False}
    if include_maybe_having_request_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.request_body is Maybe}
    if include_having_response_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.response_body is True}
    if include_not_having_response_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.response_body is False}
    if include_maybe_having_response_body is False:
        methods -= {name for name, method in HTTP_METHODS.items() if method.response_body is Maybe}

    methods -= set(exclude)

    return pytest.mark.parametrize("method", list(sorted(methods)))



import logging
import sys

from asyncpg import PostgresError
from hypercorn.utils import LifespanFailureError

from harp import get_logger

logger = get_logger(__name__)


class HypercornAdapter:
    def __init__(self, factory):
        self.factory = factory

    def _create_config(self, binds):
        """
        Creates a hypercorn config object.

        :return: hypercorn config object
        """
        from hypercorn.config import Config

        config = Config()
        config.bind = [*map(str, binds)]
        config.accesslog = logging.getLogger("hypercorn.access")
        config.errorlog = logging.getLogger("hypercorn.error")
        return config

    async def serve(self):
        """
        Creates and serves the proxy using hypercorn.
        """
        from hypercorn.asyncio import serve

        asgi_app, binds = await self.factory.build()
        config = self._create_config(binds)
        logger.debug(f"🌎 {type(self).__name__}::serve({', '.join(config.bind)})")

        try:
            return await serve(asgi_app, config)
        except LifespanFailureError as exc:
            logger.exception(f"Server initiliation failed: {repr(exc.__cause__)}", exc_info=exc.__cause__)
            if isinstance(exc.__cause__, PostgresError):
                logger.error("Could not connect to underlying postgres storage backend, check your config.")

            sys.exit(-1)



import os

from .base import BaseSetting, settings_dataclass


@settings_dataclass
class FromFileSetting(BaseSetting):
    from_file: str

    @classmethod
    def may_override(cls, instance, attr):
        _value = getattr(instance, attr)
        if not isinstance(_value, dict) or tuple(_value.keys()) != ("fromFile",):
            return
        object.__setattr__(instance, attr, FromFileSetting(from_file=_value["fromFile"]))

    def exists(self):
        return os.path.exists(self.from_file) and os.path.isfile(self.from_file)

    def open(self, *args, **kwargs):
        return open(self.from_file, *args, **kwargs)



import pytest

from harp.http.requests import WrappedHttpRequest
from harp.http.tests.test_requests import BaseHttpRequestTest


class TestWrappedHttpRequest(BaseHttpRequestTest):
    def create_request_and_wrapped_request(self, **kwargs):
        request = self.create_request(**kwargs)
        wrapped = WrappedHttpRequest(request)
        return request, wrapped

    def test_basics(self):
        initial_headers = {
            "host": "localhost",
            "connection": "keep-alive",
        }
        request, wrapped = self.create_request_and_wrapped_request(headers=initial_headers)

        # override one header
        wrapped.headers["host"] = "example.com"

        # original headers were not modified (key by key, and by full value)
        assert request.headers["host"] == "localhost"
        assert request.headers["connection"] == "keep-alive"
        assert request.headers == initial_headers

        # wrapped headers contains our overrides, and the untouched values from the original request
        assert wrapped.headers["host"] == "example.com"
        assert wrapped.headers["connection"] == "keep-alive"
        assert wrapped.headers == initial_headers | {"host": "example.com"}

    def test_delete_not_set(self):
        initial_headers = {
            "host": "localhost",
            "connection": "keep-alive",
        }
        request, wrapped = self.create_request_and_wrapped_request(headers=initial_headers)

        with pytest.raises(TypeError):
            del request.headers["foo"]

        assert request.headers == initial_headers
        assert list(request.headers.items()) == list(initial_headers.items())

        with pytest.raises(KeyError):
            del wrapped.headers["foo"]

        assert request.headers == initial_headers
        assert list(request.headers.items()) == list(initial_headers.items())

    def test_delete_set(self):
        initial_headers = {
            "host": "localhost",
            "connection": "keep-alive",
        }
        request, wrapped = self.create_request_and_wrapped_request(headers=initial_headers)

        del wrapped.headers["host"]
        assert "host" not in wrapped.headers
        assert "host" in request.headers
        assert request.headers["host"] == "localhost"
        assert dict(wrapped.headers) == {"connection": "keep-alive"}
        assert list(wrapped.headers.items()) == [("connection", "keep-alive")]
        assert wrapped.headers == {"connection": "keep-alive"}

    def test_pop(self):
        initial_headers = {
            "host": "localhost",
            "connection": "keep-alive",
        }
        request, wrapped = self.create_request_and_wrapped_request(headers=initial_headers)

        assert wrapped.headers.pop("host") == "localhost"
        assert "host" not in wrapped.headers
        assert "host" in request.headers
        assert request.headers["host"] == "localhost"
        assert wrapped.headers == {"connection": "keep-alive"}
        assert list(wrapped.headers.items()) == [("connection", "keep-alive")]

        wrapped.headers["host"] = "example.com"
        assert wrapped.headers.pop("host") == "example.com"
        assert "host" not in wrapped.headers



"""
Dashboard Application

todo:

remove self ?
remove repetition of configuration namespace ?

"""

from harp import get_logger
from harp.config import Application
from harp.config.events import FactoryBindEvent, FactoryBoundEvent

from .controllers import DashboardController
from .settings import DashboardSettings

logger = get_logger(__name__)


class DashboardApplication(Application):
    settings_namespace = "dashboard"
    settings_type = DashboardSettings

    @classmethod
    def defaults(cls, settings=None) -> dict:
        settings = settings if settings is not None else {}
        settings.setdefault("port", 4080)
        settings.setdefault("auth", None)
        return settings

    async def on_bind(self, event: FactoryBindEvent):
        event.container.register(DashboardController)
        event.container.add_alias("dashboard.controller", DashboardController)

    async def on_bound(self, event: FactoryBoundEvent):
        # add our controller to the controller resolver
        controller = event.provider.get("dashboard.controller")
        event.resolver.add(self.settings.port, controller)



from decimal import Decimal
from statistics import StatisticsError, mean

from harp.controllers import GetHandler, RouterPrefix, RoutingController
from harp.http import HttpRequest
from harp.typing.storage import Storage
from harp.views import json
from harp_apps.sqlalchemy_storage.constants import TimeBucket

from ..utils.dates import generate_continuous_time_range, get_start_datetime_from_range

time_bucket_for_range = {
    "1h": TimeBucket.HOUR.value,
    "24h": TimeBucket.HOUR.value,
    "7d": TimeBucket.DAY.value,
    "1m": TimeBucket.DAY.value,
    "1y": TimeBucket.DAY.value,
}


def _format_aggregate(items, count, key):
    if count >= 86400:
        period = "sec"
        divider = 86400

    elif count >= 1440:
        period = "min"
        divider = 1440

    elif count >= 24:
        period = "hour"
        divider = 24

    else:
        period = "day"
        divider = 1

    return {
        "rate": Decimal(int(10 * (count / divider))) / 10,
        "period": period,
        "data": [
            {
                "value": int(t[key]) if t[key] is not None else None,
            }
            for t in items
        ],
    }


@RouterPrefix("/api/overview")
class OverviewController(RoutingController):
    def __init__(self, *, storage: Storage, handle_errors=True, router=None):
        self.storage = storage
        super().__init__(handle_errors=handle_errors, router=router)

    @GetHandler("/summary")
    async def get_summary_data(self, request: HttpRequest):
        time_span = "24h"
        time_bucket = time_bucket_for_range[time_span]
        start_datetime = get_start_datetime_from_range(time_span)
        transactions_by_date_list = await self.storage.transactions_grouped_by_time_bucket(
            start_datetime=start_datetime, time_bucket=time_bucket
        )

        errors_count = sum([t["errors"] for t in transactions_by_date_list])
        transactions_count = sum([t["count"] for t in transactions_by_date_list])

        transactions_by_date_list = generate_continuous_time_range(
            discontinuous_transactions=transactions_by_date_list, time_bucket=time_bucket, start_datetime=start_datetime
        )
        try:
            mean_apdex = mean(filter(None, [t["meanApdex"] for t in transactions_by_date_list]))
        except StatisticsError:
            mean_apdex = 100

        return json(
            {
                "apdex": {
                    "mean": int(mean_apdex),
                    "data": [
                        {
                            "value": int(t["meanApdex"]) if t["meanApdex"] is not None else None,
                        }
                        for t in transactions_by_date_list
                    ],
                },
                "transactions": _format_aggregate(transactions_by_date_list, transactions_count, "count"),
                "errors": _format_aggregate(transactions_by_date_list, errors_count, "errors"),
            }
        )

    @GetHandler("/")
    async def get_overview_data(self, request: HttpRequest):
        # endpoint and range from request
        endpoint = request.query.get("endpoint")
        range = request.query.get("timeRange", "24h")

        # time buckets and start_datetime accordingly
        time_bucket = time_bucket_for_range.get(range, "day")
        start_datetime = get_start_datetime_from_range(range)

        transactions_by_date_list = await self.storage.transactions_grouped_by_time_bucket(
            endpoint=endpoint, start_datetime=start_datetime, time_bucket=time_bucket
        )
        errors_count = sum([t["errors"] for t in transactions_by_date_list])
        transactions_count = sum([t["count"] for t in transactions_by_date_list])
        errors_rate = errors_count / transactions_count if transactions_count else 0
        mean_duration = (
            sum([t["meanDuration"] * t["count"] for t in transactions_by_date_list]) / transactions_count
            if transactions_count
            else 0
        )
        transactions_by_date_list = generate_continuous_time_range(
            discontinuous_transactions=transactions_by_date_list, time_bucket=time_bucket, start_datetime=start_datetime
        )

        try:
            mean_apdex = mean(filter(None, [t["meanApdex"] for t in transactions_by_date_list]))
        except StatisticsError:
            mean_apdex = 100

        return json(
            {
                "transactions": transactions_by_date_list,
                "errors": {"count": errors_count, "rate": errors_rate},
                "count": transactions_count,
                "meanDuration": mean_duration,
                "meanApdex": mean_apdex,
                "timeRange": range,
            }
        )



from typing import Type, cast

from rodi import Container
from whistle import IAsyncEventDispatcher

from harp import __revision__, __version__, get_logger
from harp.asgi import ASGIKernel
from harp.asgi.events import EVENT_CONTROLLER_VIEW, EVENT_CORE_REQUEST, RequestEvent
from harp.config import Config
from harp.config.events import (
    EVENT_FACTORY_BIND,
    EVENT_FACTORY_BOUND,
    EVENT_FACTORY_BUILD,
    FactoryBindEvent,
    FactoryBoundEvent,
    FactoryBuildEvent,
)
from harp.controllers import ProxyControllerResolver
from harp.event_dispatcher import LoggingAsyncEventDispatcher
from harp.http import HttpResponse
from harp.typing import GlobalSettings
from harp.utils.network import Bind
from harp.views.json import on_json_response

logger = get_logger(__name__)


async def ok_controller():
    return HttpResponse("Ok.", status=200)


async def on_health_request(event: RequestEvent):
    if event.request.path == "/healthz":
        event.set_controller(ok_controller)
        event.stop_propagation()


class KernelFactory:
    AsyncEventDispatcherType: Type[IAsyncEventDispatcher] = LoggingAsyncEventDispatcher
    ContainerType: Type[Container] = Container
    KernelType: Type[ASGIKernel] = ASGIKernel

    def __init__(self, configuration: Config):
        self.configuration = configuration
        self.hostname = "[::]"

    async def build(self):
        logger.info(f"🎙  HARP v.{__version__} ({__revision__})")
        # we only work on validated configuration
        self.configuration.validate()

        logger.info(f"📦 Apps: {', '.join(self.configuration.applications)}")

        dispatcher = self.build_event_dispatcher()
        container = self.build_container(dispatcher)
        resolver = ProxyControllerResolver()

        # dispatch "bind" event: this is the last chance to add services to the container
        try:
            await dispatcher.adispatch(
                EVENT_FACTORY_BIND,
                FactoryBindEvent(
                    container,
                    self.configuration.settings,
                ),
            )
        except Exception as exc:
            logger.fatal("Fatal while dispatching «factory bind» event: %s", exc)
            raise

        # this can fail if configuration is not valid, but we let the container raise exception which is more explicit
        # that what we can show here.
        provider = container.build_provider()

        # dispatch "bound" event: you get a resolved container, do your thing
        try:
            await dispatcher.adispatch(
                EVENT_FACTORY_BOUND,
                FactoryBoundEvent(
                    provider,
                    resolver,
                ),
            )
        except Exception as exc:
            logger.fatal("Fatal while dispatching «factory bound» event: %s", exc)
            raise

        kernel = self.KernelType(dispatcher=dispatcher, resolver=resolver)
        binds = [Bind(host=self.hostname, port=port) for port in resolver.ports]
        event = FactoryBuildEvent(kernel, binds)
        await dispatcher.adispatch(EVENT_FACTORY_BUILD, event)

        return event.kernel, event.binds

    def build_container(self, dispatcher: IAsyncEventDispatcher) -> Container:
        """Factory method responsible for the service injection container creation, registering initial services."""
        container = cast(Container, self.ContainerType())
        container.add_instance(self.configuration.settings, GlobalSettings)
        container.add_instance(dispatcher, IAsyncEventDispatcher)

        self.configuration.register_services(container)

        return container

    def build_event_dispatcher(self) -> IAsyncEventDispatcher:
        """Factory method responsible for the event dispatcher creation, binding initial/generic listeners."""
        dispatcher = cast(IAsyncEventDispatcher, self.AsyncEventDispatcherType())

        dispatcher.add_listener(EVENT_CORE_REQUEST, on_health_request, priority=-100)

        # todo move into core or extension, this is not proxy related
        dispatcher.add_listener(EVENT_CONTROLLER_VIEW, on_json_response)

        self.configuration.register_events(dispatcher)

        return dispatcher



from dataclasses import field
from typing import TYPE_CHECKING, Optional

from hishel import HEURISTICALLY_CACHEABLE_STATUS_CODES

if TYPE_CHECKING:
    from hishel import AsyncCacheTransport, Controller
    from hishel._async._storages import AsyncBaseStorage
    from httpx import AsyncHTTPTransport

from harp.config import BaseSetting, Definition, DisableableBaseSettings, Lazy, settings_dataclass
from harp.settings import DEFAULT_TIMEOUT


@settings_dataclass
class CacheSettings(DisableableBaseSettings):
    transport: Definition["AsyncCacheTransport"] = Lazy("hishel:AsyncCacheTransport")
    controller: Definition["Controller"] = Lazy(
        "hishel:Controller",
        allow_heuristics=True,
        allow_stale=True,
        cacheable_methods=["GET", "HEAD"],
        cacheable_status_codes=list(HEURISTICALLY_CACHEABLE_STATUS_CODES),
    )
    storage: Definition["AsyncBaseStorage"] = Lazy(None)


@settings_dataclass
class HttpClientSettings(BaseSetting):
    timeout: Optional[float] = DEFAULT_TIMEOUT
    cache: CacheSettings = field(default_factory=CacheSettings)

    #: HTTP transport to use for the client. This is usually a httpx.AsyncHTTPTransport (or subclass) instance.
    transport: Definition["AsyncHTTPTransport"] = Lazy("httpx:AsyncHTTPTransport")

    def __post_init__(self):
        super().__post_init__()

        if self.cache and isinstance(self.cache, dict):
            self.cache = CacheSettings(**self.cache)



import os
from tempfile import NamedTemporaryFile

from harp.config.settings import DisabledSettings, FromFileSetting


def test_from_file_setting_not_existing():
    filename = "/this/is/very/unlikely/to/exist"
    setting = FromFileSetting(from_file=filename)
    assert setting.to_dict() == {"from_file": filename}
    assert not setting.exists()


def test_from_file_setting_existing():
    try:
        with NamedTemporaryFile("w+", delete=False) as tmp_file:
            tmp_file.write("test")
            setting = FromFileSetting(from_file=tmp_file.name)
            assert setting.to_dict() == {"from_file": tmp_file.name}
            assert setting.exists()

        with setting.open() as f:
            assert f.read() == "test"
    finally:
        if os.path.exists(tmp_file.name):
            os.unlink(tmp_file.name)


def test_disabled_settings():
    setting = DisabledSettings()
    assert setting.to_dict() == {"enabled": False}
    assert repr(setting) == "disabled"



from typing import Optional

from harp.config.settings.base import BaseSetting, settings_dataclass


@settings_dataclass
class ProxyEndpointSetting(BaseSetting):
    name: str
    port: int
    url: str

    description: Optional[str] = None


@settings_dataclass
class ProxySettings(BaseSetting):
    endpoints: Optional[list[ProxyEndpointSetting]]

    def __post_init__(self):
        if self.endpoints is None:
            self.endpoints = []

        self.endpoints = [
            endpoint if isinstance(endpoint, ProxyEndpointSetting) else ProxyEndpointSetting(**endpoint)
            for endpoint in self.endpoints
        ]



from .base import BaseSetting, asdict, settings_dataclass
from .disabled import DisableableBaseSettings, DisabledSettings
from .from_file import FromFileSetting
from .lazy import Definition, Lazy

__all__ = [
    "BaseSetting",
    "Definition",
    "DisableableBaseSettings",
    "DisabledSettings",
    "FromFileSetting",
    "Lazy",
    "asdict",
    "settings_dataclass",
]
